from enum import IntEnum
from pathlib import Path
from typing import Iterable, Sequence

import numpy as np

from _common.types import ColorLike, Vector2Like
from .._common import to_rgba
from .._viewer_protocol import IViewer
from .._core.image import Image
from .._core.polyline import PolyLine
from .bbox import BBox2


class PolyContour:
    class EWinding(IntEnum):
        UNKNOWN = 0
        CLOCKWISE = 1
        COUNTERCLOCKWISE = 2

    @staticmethod
    def strWindingAsString(eWinding: "PolyContour.EWinding") -> str:
        if eWinding == PolyContour.EWinding.COUNTERCLOCKWISE:
            return "[counter-clockwise]"
        if eWinding == PolyContour.EWinding.CLOCKWISE:
            return "[clockwise]"
        return "[unknown/degenerate]"

    def __init__(self, vertices: Iterable[Vector2Like], eWinding: "PolyContour.EWinding" = EWinding.UNKNOWN) -> None:
        self._vertices = [(float(v[0]), float(v[1])) for v in vertices]
        if len(self._vertices) < 3:
            raise ValueError("Polyline with less than 3 points makes no sense")
        self._bbox = BBox2()
        for v in self._vertices:
            self._bbox.Include(v)
        self._winding = self.eDetectWinding(self._vertices) if eWinding == self.EWinding.UNKNOWN else eWinding

    @staticmethod
    def eDetectWinding(vertices: Sequence[Vector2Like]) -> "PolyContour.EWinding":
        if len(vertices) < 3:
            return PolyContour.EWinding.UNKNOWN
        area = 0.0
        for i in range(len(vertices)):
            j = (i + 1) % len(vertices)
            area += (vertices[j][0] - vertices[i][0]) * (vertices[j][1] + vertices[i][1])
        if area > 0:
            return PolyContour.EWinding.CLOCKWISE
        if area < 0:
            return PolyContour.EWinding.COUNTERCLOCKWISE
        return PolyContour.EWinding.UNKNOWN

    def AddVertex(self, vec: Vector2Like) -> None:
        v = (float(vec[0]), float(vec[1]))
        self._vertices.append(v)
        self._bbox.Include(v)

    def DetectWinding(self) -> None:
        self._winding = self.eDetectWinding(self._vertices)

    def eWinding(self) -> "PolyContour.EWinding":
        return self._winding

    def oVertices(self) -> list[tuple[float, float]]:
        return list(self._vertices)

    def Close(self) -> None:
        if not self._vertices:
            return
        if self._vertices[0] != self._vertices[-1]:
            self._vertices.append(self._vertices[0])

    def AsSvgPolyline(self) -> str:
        pts = " ".join(f"{x},{y}" for x, y in self._vertices + [self._vertices[0]])
        stroke = "red"
        if self._winding == self.EWinding.CLOCKWISE:
            stroke = "blue"
        elif self._winding == self.EWinding.COUNTERCLOCKWISE:
            stroke = "black"
        return f"<polyline points='{pts}' stroke='{stroke}' fill='none' stroke-width='0.1' />\n"

    def AsSvgPath(self) -> str:
        parts: list[str] = []
        for i, (x, y) in enumerate(self._vertices):
            if i == 0:
                parts.append(f"M{x},{y}")
            else:
                parts.append(f"L{x},{y}")
        parts.append("Z")
        return " ".join(parts)

    def oBBox(self) -> BBox2:
        return self._bbox

    def nCount(self) -> int:
        return len(self._vertices)

    def vecVertex(self, n: int) -> tuple[float, float]:
        return self._vertices[n]


class PolySlice:
    def __init__(self, fZPos: float) -> None:
        self.m_fZPos = float(fZPos)
        self._contours: list[PolyContour] = []
        self._bbox = BBox2()

    def fZPos(self) -> float:
        return self.m_fZPos

    def AddContour(self, contour: PolyContour) -> None:
        self._contours.append(contour)
        self._bbox.Include(contour.oBBox())

    def nContours(self) -> int:
        return len(self._contours)

    def oContourAt(self, index: int) -> PolyContour:
        return self._contours[index]

    def oBBox(self) -> BBox2:
        return self._bbox

    def bIsEmpty(self) -> bool:
        return len(self._contours) == 0

    def Close(self) -> None:
        for c in self._contours:
            c.Close()

    def SaveToSvgFile(self, strPath: str | Path, bSolid: bool, oBBoxToUse: BBox2 | None = None) -> None:
        bbox = oBBoxToUse or self._bbox
        sx = bbox.vecMax[0] - bbox.vecMin[0]
        sy = bbox.vecMax[1] - bbox.vecMin[1]
        with open(strPath, "w", encoding="utf-8") as handle:
            handle.write("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n")
            handle.write("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n")
            handle.write(
                f"<svg xmlns='http://www.w3.org/2000/svg' version='1.1' viewBox='{bbox.vecMin[0]} {bbox.vecMin[1]} {sx} {sy}' width='{sx}mm' height='{sy}mm'>\n"
            )
            handle.write("<g>\n")
            if not bSolid:
                for c in self._contours:
                    handle.write(c.AsSvgPolyline())
            else:
                path = " ".join(c.AsSvgPath() for c in self._contours)
                handle.write(f"<path d='{path}' fill='black'/>\n")
            handle.write("</g>\n</svg>\n")

    @staticmethod
    def _zero_crossing(a: float, b: float) -> float:
        den = a - b
        if abs(den) < 1e-8:
            return 0.5
        return float(np.clip(a / den, 0.0, 1.0))

    @staticmethod
    def oFromSdf(img: Image, fZPos: float, vecOffset: Vector2Like, fScale: float) -> "PolySlice":
        slice_obj = PolySlice(fZPos)
        if img.nWidth < 2 or img.nHeight < 2:
            return slice_obj
        ox, oy = float(vecOffset[0]), float(vecOffset[1])

        segs: list[tuple[tuple[float, float], tuple[float, float]]] = []
        edge_map = {
            1: ((3, 0),),
            2: ((0, 1),),
            3: ((3, 1),),
            4: ((1, 2),),
            5: ((3, 2), (0, 1)),
            6: ((0, 2),),
            7: ((3, 2),),
            8: ((2, 3),),
            9: ((0, 2),),
            10: ((0, 1), (2, 3)),
            11: ((1, 2),),
            12: ((1, 3),),
            13: ((0, 1),),
            14: ((0, 3),),
        }
        for y in range(img.nHeight - 1):
            for x in range(img.nWidth - 1):
                corners = [img.fValue(x, y), img.fValue(x + 1, y), img.fValue(x + 1, y + 1), img.fValue(x, y + 1)]
                idx = 0
                if corners[0] < 0.0:
                    idx |= 1
                if corners[1] < 0.0:
                    idx |= 2
                if corners[2] < 0.0:
                    idx |= 4
                if corners[3] < 0.0:
                    idx |= 8
                if idx == 0 or idx == 15:
                    continue
                cross: dict[int, tuple[float, float]] = {}
                if idx in edge_map:
                    if any(e in (0, 1, 2, 3) for pair in edge_map[idx] for e in pair):
                        cross[0] = (x + PolySlice._zero_crossing(corners[0], corners[1]), y)
                        cross[1] = (x + 1, y + PolySlice._zero_crossing(corners[1], corners[2]))
                        cross[2] = (x + PolySlice._zero_crossing(corners[3], corners[2]), y + 1)
                        cross[3] = (x, y + PolySlice._zero_crossing(corners[0], corners[3]))
                    for e0, e1 in edge_map[idx]:
                        p0 = cross[e0]
                        p1 = cross[e1]
                        segs.append(((ox + p0[0] * fScale, oy + p0[1] * fScale), (ox + p1[0] * fScale, oy + p1[1] * fScale)))

        remaining = segs[:]
        while remaining:
            a, b = remaining.pop()
            contour = [a, b]
            changed = True
            while changed and remaining:
                changed = False
                for i, (s0, s1) in enumerate(remaining):
                    if np.hypot(contour[-1][0] - s0[0], contour[-1][1] - s0[1]) < 1e-3:
                        contour.append(s1)
                        remaining.pop(i)
                        changed = True
                        break
                    if np.hypot(contour[-1][0] - s1[0], contour[-1][1] - s1[1]) < 1e-3:
                        contour.append(s0)
                        remaining.pop(i)
                        changed = True
                        break
            if len(contour) >= 3:
                try:
                    slice_obj.AddContour(PolyContour(contour))
                except ValueError:
                    pass
        return slice_obj


class PolySliceStack:
    def __init__(self, slices: Sequence[PolySlice] | None = None) -> None:
        self._slices = list(slices or [])
        self._bbox = BBox2()
        for sl in self._slices:
            self._bbox.Include(sl.oBBox())

    def AddSlice(self, poly_slice: PolySlice) -> None:
        self._slices.append(poly_slice)
        self._bbox.Include(poly_slice.oBBox())

    def AddSlices(self, slices: Sequence[PolySlice]) -> None:
        for sl in slices:
            self.AddSlice(sl)

    def nCount(self) -> int:
        return len(self._slices)

    def oSliceAt(self, index: int) -> PolySlice:
        return self._slices[index]

    def oBBox(self) -> BBox2:
        return self._bbox

    def oSlices(self) -> list[PolySlice]:
        return list(self._slices)

    def AddToViewer(
        self,
        oViewer: IViewer,
        clrOutside: ColorLike | None = None,
        clrInside: ColorLike | None = None,
        clrDegenerate: ColorLike | None = None,
        nGroup: int = 0,
    ) -> None:
        degenerate = to_rgba(clrDegenerate or (2.0 / 3.0, 2.0 / 3.0, 2.0 / 3.0, 2.0 / 3.0))
        inside = to_rgba(clrInside or (2.0 / 3.0, 2.0 / 3.0, 2.0 / 3.0, 2.0 / 3.0))
        outside = to_rgba(clrOutside or (1.0, 0.0, 0.0, 2.0 / 3.0))

        add_fn = getattr(oViewer, "Add", None)
        if not callable(add_fn):
            raise TypeError("oViewer must expose an Add method")

        for sl in self._slices:
            for idx in range(sl.nContours()):
                contour = sl.oContourAt(idx)

                color = degenerate
                if contour.eWinding() == PolyContour.EWinding.CLOCKWISE:
                    color = inside
                elif contour.eWinding() == PolyContour.EWinding.COUNTERCLOCKWISE:
                    color = outside

                poly = PolyLine(color)
                for vec in contour.oVertices():
                    poly.nAddVertex((vec[0], vec[1], sl.fZPos()))

                add_fn(poly, int(nGroup))





from enum import IntEnum
import math


class Easing:
    class EEasing(IntEnum):
        LINEAR = 0
        SINE_IN = 1
        SINE_OUT = 2
        SINE_INOUT = 3
        QUAD_IN = 4
        QUAD_OUT = 5
        QUAD_INOUT = 6
        CUBIC_IN = 7
        CUBIC_OUT = 8
        CUBIC_INOUT = 9

    @staticmethod
    def fEaseSineIn(x: float) -> float:
        return 1.0 - math.cos((x * math.pi) / 2.0)

    @staticmethod
    def fEaseSineOut(x: float) -> float:
        return math.sin((x * math.pi) / 2.0)

    @staticmethod
    def fEaseSineInOut(x: float) -> float:
        return -(math.cos(math.pi * x) - 1.0) / 2.0

    @staticmethod
    def fEaseQuadIn(x: float) -> float:
        return x * x

    @staticmethod
    def fEaseQuadOut(x: float) -> float:
        return 1.0 - (1.0 - x) * (1.0 - x)

    @staticmethod
    def fEaseQuadInOut(x: float) -> float:
        return 2.0 * x * x if x < 0.5 else 1.0 - ((-2.0 * x + 2.0) ** 2.0) / 2.0

    @staticmethod
    def fEaseCubicIn(x: float) -> float:
        return x * x * x

    @staticmethod
    def fEaseCubicOut(x: float) -> float:
        return 1.0 - ((1.0 - x) ** 3.0)

    @staticmethod
    def fEaseCubicInOut(x: float) -> float:
        return 4.0 * x * x * x if x < 0.5 else 1.0 - ((-2.0 * x + 2.0) ** 3.0) / 2.0

    @staticmethod
    def fEasingFunction(x: float, easing: "Easing.EEasing") -> float:
        mapping = {
            Easing.EEasing.LINEAR: lambda y: y,
            Easing.EEasing.SINE_IN: Easing.fEaseSineIn,
            Easing.EEasing.SINE_OUT: Easing.fEaseSineOut,
            Easing.EEasing.SINE_INOUT: Easing.fEaseSineInOut,
            Easing.EEasing.QUAD_IN: Easing.fEaseQuadIn,
            Easing.EEasing.QUAD_OUT: Easing.fEaseQuadOut,
            Easing.EEasing.QUAD_INOUT: Easing.fEaseQuadInOut,
            Easing.EEasing.CUBIC_IN: Easing.fEaseCubicIn,
            Easing.EEasing.CUBIC_OUT: Easing.fEaseCubicOut,
            Easing.EEasing.CUBIC_INOUT: Easing.fEaseCubicInOut,
        }
        try:
            return float(mapping[easing](x))
        except KeyError as exc:
            raise ValueError("Unknown easing function") from exc




from __future__ import annotations

from picogk import Lattice, Library, Voxels
from shape_kernel import LocalFrame

from .icosahedral_face import IcosehedralFace
from .quasi_tile import QuasiTile
from .quasi_tile_01 import QuasiTile_01
from .quasi_tile_04 import QuasiTile_04
from .quasi_tile_inflation import QuasiTileInflation


class QuasiCrystal:
    def __init__(
        self,
        nGenerations: int,
        aInitialTiles: list[QuasiTile] | None = None,
        sInitialFace: IcosehedralFace | None = None,
    ) -> None:
        self.m_nGenerations = int(nGenerations)
        self.m_aTileGenerations: list[list[QuasiTile]] = [[] for _ in range(self.m_nGenerations)]

        if aInitialTiles is not None:
            self.m_aTileGenerations[0] = list(aInitialTiles)
        elif sInitialFace is not None:
            self.m_aTileGenerations[0] = QuasiTileInflation.aGetInflatedFace(sInitialFace)
        else:
            raise ValueError("Provide either aInitialTiles or sInitialFace.")

        for i in range(1, self.m_nGenerations):
            a_inflated_sub_tiles: list[QuasiTile] = []
            for o_tile in self.m_aTileGenerations[i - 1]:
                a_faces = o_tile.aGetFaces()
                for s_face in a_faces:
                    a_inflated_sub_tiles.extend(QuasiTileInflation.aGetInflatedFace(s_face))

            a_inflated_sub_tiles = self.aGetDeduplicatedSubTiles(a_inflated_sub_tiles)
            self.m_aTileGenerations[i] = a_inflated_sub_tiles

    def aGetDeduplicatedSubTiles(self, aSubTiles: list[QuasiTile]) -> list[QuasiTile]:
        a_ref_centres: list[tuple[float, float, float]] = []
        a_deduplicated_sub_tiles: list[QuasiTile] = []

        for o_sub_tile in aSubTiles:
            vec_rounded_centre = o_sub_tile.vecGetRoundedCentre()
            if vec_rounded_centre not in a_ref_centres:
                a_ref_centres.append(vec_rounded_centre)
                a_deduplicated_sub_tiles.append(o_sub_tile)

        return a_deduplicated_sub_tiles

    @staticmethod
    def aGetFirstGenerationTiles() -> list[QuasiTile]:
        a_tiles: list[QuasiTile] = []

        o_tile_000 = QuasiTile_01(LocalFrame())
        o_tile_001 = QuasiTile_01(LocalFrame())
        o_tile_002 = QuasiTile_01(LocalFrame())
        o_tile_003 = QuasiTile_01(LocalFrame())
        o_tile_004 = QuasiTile_01(LocalFrame())

        o_tile_001.AttachToOtherQuasiTile(0, o_tile_000, 1)
        o_tile_002.AttachToOtherQuasiTile(0, o_tile_001, 1)
        o_tile_003.AttachToOtherQuasiTile(0, o_tile_002, 1)
        o_tile_004.AttachToOtherQuasiTile(1, o_tile_000, 0)

        o_tile_005 = QuasiTile_01(LocalFrame())
        o_tile_006 = QuasiTile_01(LocalFrame())
        o_tile_007 = QuasiTile_01(LocalFrame())
        o_tile_008 = QuasiTile_01(LocalFrame())
        o_tile_009 = QuasiTile_01(LocalFrame())

        o_tile_005.AttachToOtherQuasiTile(0, o_tile_000, 2)
        o_tile_006.AttachToOtherQuasiTile(0, o_tile_001, 2)
        o_tile_007.AttachToOtherQuasiTile(0, o_tile_002, 2)
        o_tile_008.AttachToOtherQuasiTile(0, o_tile_003, 2)
        o_tile_009.AttachToOtherQuasiTile(0, o_tile_004, 2)

        o_tile_010 = QuasiTile_01(LocalFrame())
        o_tile_011 = QuasiTile_01(LocalFrame())
        o_tile_012 = QuasiTile_01(LocalFrame())
        o_tile_013 = QuasiTile_01(LocalFrame())
        o_tile_014 = QuasiTile_01(LocalFrame())

        o_tile_010.AttachToOtherQuasiTile(0, o_tile_005, 1)
        o_tile_011.AttachToOtherQuasiTile(0, o_tile_006, 1)
        o_tile_012.AttachToOtherQuasiTile(0, o_tile_007, 1)
        o_tile_013.AttachToOtherQuasiTile(0, o_tile_008, 1)
        o_tile_014.AttachToOtherQuasiTile(0, o_tile_009, 1)

        o_tile_015 = QuasiTile_01(LocalFrame())
        o_tile_016 = QuasiTile_01(LocalFrame())
        o_tile_017 = QuasiTile_01(LocalFrame())
        o_tile_018 = QuasiTile_01(LocalFrame())
        o_tile_019 = QuasiTile_01(LocalFrame())

        o_tile_015.AttachToOtherQuasiTile(0, o_tile_010, 2)
        o_tile_016.AttachToOtherQuasiTile(0, o_tile_011, 2)
        o_tile_017.AttachToOtherQuasiTile(0, o_tile_012, 2)
        o_tile_018.AttachToOtherQuasiTile(0, o_tile_013, 2)
        o_tile_019.AttachToOtherQuasiTile(0, o_tile_014, 2)

        a_tiles.extend([
            o_tile_000,
            o_tile_001,
            o_tile_002,
            o_tile_003,
            o_tile_004,
            o_tile_005,
            o_tile_006,
            o_tile_007,
            o_tile_008,
            o_tile_009,
            o_tile_010,
            o_tile_011,
            o_tile_012,
            o_tile_013,
            o_tile_014,
            o_tile_015,
            o_tile_016,
            o_tile_017,
            o_tile_018,
            o_tile_019,
        ])
        return a_tiles

    @staticmethod
    def aGetSecondGenerationTiles() -> list[QuasiTile]:
        a_first_generation_tiles = QuasiCrystal.aGetFirstGenerationTiles()
        a_tiles: list[QuasiTile] = []

        o_tile_100 = QuasiTile_04(LocalFrame())
        o_tile_101 = QuasiTile_04(LocalFrame())
        o_tile_102 = QuasiTile_04(LocalFrame())
        o_tile_103 = QuasiTile_04(LocalFrame())
        o_tile_104 = QuasiTile_04(LocalFrame())
        o_tile_105 = QuasiTile_04(LocalFrame())
        o_tile_106 = QuasiTile_04(LocalFrame())
        o_tile_107 = QuasiTile_04(LocalFrame())
        o_tile_108 = QuasiTile_04(LocalFrame())
        o_tile_109 = QuasiTile_04(LocalFrame())
        o_tile_110 = QuasiTile_04(LocalFrame())
        o_tile_111 = QuasiTile_04(LocalFrame())

        o_tile_100.AttachToOtherQuasiTile(1, a_first_generation_tiles[0], 4)
        o_tile_101.AttachToOtherQuasiTile(1, a_first_generation_tiles[0], 5)
        o_tile_102.AttachToOtherQuasiTile(1, a_first_generation_tiles[1], 5)
        o_tile_103.AttachToOtherQuasiTile(1, a_first_generation_tiles[2], 5)
        o_tile_104.AttachToOtherQuasiTile(1, a_first_generation_tiles[3], 5)
        o_tile_105.AttachToOtherQuasiTile(1, a_first_generation_tiles[4], 5)
        o_tile_106.AttachToOtherQuasiTile(1, a_first_generation_tiles[15], 5)
        o_tile_107.AttachToOtherQuasiTile(1, a_first_generation_tiles[16], 5)
        o_tile_108.AttachToOtherQuasiTile(1, a_first_generation_tiles[17], 5)
        o_tile_109.AttachToOtherQuasiTile(1, a_first_generation_tiles[18], 5)
        o_tile_110.AttachToOtherQuasiTile(1, a_first_generation_tiles[19], 5)
        o_tile_111.AttachToOtherQuasiTile(1, a_first_generation_tiles[19], 3)

        a_tiles.extend([
            o_tile_100,
            o_tile_101,
            o_tile_102,
            o_tile_103,
            o_tile_104,
            o_tile_105,
            o_tile_106,
            o_tile_107,
            o_tile_108,
            o_tile_109,
            o_tile_110,
            o_tile_111,
        ])
        return a_tiles

    def PreviewGeneration(self, nGen: int, ePreviewFace: QuasiTile.EPreviewFace) -> None:
        try:
            a_tiles = self.m_aTileGenerations[int(nGen)]
            Library.Log(f"Number of Tiles = {len(a_tiles)}.")
            for o_tile in a_tiles:
                o_tile.Preview(ePreviewFace)
        except Exception as exc:
            raise Exception("Generation not found.") from exc

    def aGetTileGeneration(self, nGen: int) -> list[QuasiTile]:
        try:
            return self.m_aTileGenerations[int(nGen)]
        except Exception as exc:
            raise Exception("Generation not found.") from exc

    def voxGetWireframe(self, nGen: int, fBeamR: float) -> Voxels:
        try:
            a_quasi_tiles = self.m_aTileGenerations[int(nGen)]
            o_lattice = Lattice()

            for o_tile in a_quasi_tiles:
                for s_face in o_tile.aGetFaces():
                    vec_pt1 = s_face.vecPt1
                    vec_pt2 = s_face.vecPt2
                    vec_pt3 = s_face.vecPt3
                    vec_pt4 = s_face.vecPt4

                    o_lattice.AddBeam(vec_pt1, vec_pt2, fBeamR, fBeamR, True)
                    o_lattice.AddBeam(vec_pt2, vec_pt3, fBeamR, fBeamR, True)
                    o_lattice.AddBeam(vec_pt3, vec_pt4, fBeamR, fBeamR, True)
                    o_lattice.AddBeam(vec_pt4, vec_pt1, fBeamR, fBeamR, True)

            return Voxels.from_lattice(o_lattice)
        except Exception as exc:
            raise Exception("Generation not found.") from exc


__all__ = ["QuasiCrystal"]



from __future__ import annotations

import math

from shape_kernel._types import Vector3Like, as_vec3

from picogk import IImplicit


class ImplicitSplitWallGyroid(IImplicit):
	def __init__(self, fUnitSize: float, fWallThickness: float, bSide: bool) -> None:
		self.m_fFrequencyScale = (2.0 * math.pi) / float(fUnitSize)
		self.m_fWallThickness = float(fWallThickness)
		self.m_bSide = bool(bSide)

	def fSignedDistance(self, vecPt: Vector3Like) -> float:
		dX, dY, dZ = as_vec3(vecPt)
		dDist = (
			math.sin(self.m_fFrequencyScale * dX) * math.cos(self.m_fFrequencyScale * dY)
			+ math.sin(self.m_fFrequencyScale * dY) * math.cos(self.m_fFrequencyScale * dZ)
			+ math.sin(self.m_fFrequencyScale * dZ) * math.cos(self.m_fFrequencyScale * dX)
		)
		if self.m_bSide:
			return max(dDist, abs(dDist) - 0.5 * self.m_fWallThickness)
		return max(-dDist, abs(dDist) - 0.5 * self.m_fWallThickness)


__all__ = ["ImplicitSplitWallGyroid"]


from __future__ import annotations

from collections.abc import Sequence

import numpy as np
from .i_spline import ISpline

from .._types import Vec3, Vector3Like, as_np3, as_vec3, clamp01


class ControlPointSpline(ISpline):
    class EEnds:
        OPEN = "OPEN"
        CLOSED = "CLOSED"

    def __init__(self, aControlPoints: Sequence[Vector3Like], nDegree: int = 2, eEnds: str = EEnds.OPEN) -> None:
        self.m_aControlPoints = [as_np3(p) for p in aControlPoints]
        self.m_eEnds = eEnds
        self.m_fError = 1e-7

        if self.m_eEnds == self.EEnds.CLOSED and len(self.m_aControlPoints) > 1:
            if float(np.linalg.norm(self.m_aControlPoints[0] - self.m_aControlPoints[-1])) < self.m_fError:
                self.m_aControlPoints.pop()

        self.m_nDegree = int(nDegree)
        self.m_aKnot = self.aGetKnotVector()

    def aGetKnotVector(self) -> list[float]:
        n_control = len(self.m_aControlPoints)
        if self.m_eEnds == self.EEnds.CLOSED:
            for idx in range(n_control - 1):
                self.m_aControlPoints.append(self.m_aControlPoints[idx].copy())
            n_control = len(self.m_aControlPoints)

        n_knots = n_control + self.m_nDegree + 1
        n_valid_range = n_knots - self.m_nDegree - (self.m_nDegree + 1)
        d_r = 1.0 / float(n_valid_range)

        knots: list[float] = []
        for idx in range(n_knots):
            value = -(d_r * self.m_nDegree) + d_r * idx
            if self.m_eEnds == self.EEnds.OPEN:
                value = clamp01(value)
            knots.append(float(value))
        return knots

    def aGetPoints(self, nSamples: int = 500) -> list[Vec3]:
        points: list[Vec3] = []
        for idx in range(int(nSamples)):
            lr = float(idx) / float(max(1, nSamples - 1))
            points.append(self.vecGetPointAt(lr))
        return points

    def vecGetPointAt(self, fLengthRatio: float) -> Vec3:
        point = np.zeros(3, dtype=np.float64)
        for idx, control_point in enumerate(self.m_aControlPoints):
            point += self.fBaseFunc(fLengthRatio, idx, self.m_nDegree) * control_point
        return as_vec3(point)

    def fBaseFunc(self, fLengthRatio: float, nControlPoint: int, nDegree: int) -> float:
        if nDegree == 0:
            left = self.m_aKnot[nControlPoint]
            right = self.m_aKnot[nControlPoint + 1]
            if (left <= fLengthRatio < right) or (
                abs(fLengthRatio - right) < self.m_fError and abs(fLengthRatio - self.m_aKnot[-1]) < self.m_fError
            ):
                return 1.0
            return 0.0

        value = 0.0
        left_den = self.m_aKnot[nControlPoint + nDegree] - self.m_aKnot[nControlPoint]
        if abs(left_den) > self.m_fError:
            value += (
                (fLengthRatio - self.m_aKnot[nControlPoint])
                / left_den
                * self.fBaseFunc(fLengthRatio, nControlPoint, nDegree - 1)
            )

        right_den = self.m_aKnot[nControlPoint + nDegree + 1] - self.m_aKnot[nControlPoint + 1]
        if abs(right_den) > self.m_fError:
            value += (
                (self.m_aKnot[nControlPoint + nDegree + 1] - fLengthRatio)
                / right_den
                * self.fBaseFunc(fLengthRatio, nControlPoint + 1, nDegree - 1)
            )
        return float(value)




from __future__ import annotations

from picogk import Voxels
from shape_kernel._types import Vector3Like, as_vec3
from shape_kernel.utils import Uf

from .i_beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell


class GlobalFuncBeamThickness(IBeamThickness):
	def __init__(self, fMinBeamThickness: float, fMaxBeamThickness: float) -> None:
		self.m_fMinBeamThickness = float(fMinBeamThickness)
		self.m_fMaxBeamThickness = float(fMaxBeamThickness)

	def UpdateCell(self, xCell: IUnitCell) -> None:
		return None

	def fGetBeamThickness(self, vecPt: Vector3Like) -> float:
		pt = as_vec3(vecPt)
		fRatio = min(max(0.02 * float(pt[0]), 0.0), 1.0)
		return Uf.fTransFixed(self.m_fMinBeamThickness, self.m_fMaxBeamThickness, fRatio)

	def SetBoundingVoxels(self, voxBounding: Voxels) -> None:
		return None


__all__ = ["GlobalFuncBeamThickness"]


from __future__ import annotations

from picogk import Voxels
from shape_kernel._types import Vector3Like

from .i_beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell


class ConstantBeamThickness(IBeamThickness):
	def __init__(self, fBeamThickness: float) -> None:
		self.m_fBeamThickness = float(fBeamThickness)

	def fGetBeamThickness(self, vecPt: Vector3Like) -> float:
		return self.m_fBeamThickness

	def UpdateCell(self, xCell: IUnitCell) -> None:
		return None

	def SetBoundingVoxels(self, voxBounding: Voxels) -> None:
		return None


__all__ = ["ConstantBeamThickness"]


from .line_modulation import Distribution, GenericContour, LineModulation
from .surface_modulation import SurfaceModulation

__all__ = ['Distribution','GenericContour','LineModulation','SurfaceModulation']




from __future__ import annotations

from typing import TYPE_CHECKING

from picogk import Voxels

from .._types import as_np3, as_vec3

if TYPE_CHECKING:
    from ..frames.local_frames import LocalFrame


class CylUtility:
    @staticmethod
    def voxGetCyl(fStartZ_or_frame: float | LocalFrame, fEndZ_or_startZ: float = 0.0, fRadius_or_endZ: float = 0.0, fRadius: float | None = None) -> Voxels:
        from ..base_shapes import BaseCylinder
        from ..frames.local_frames import LocalFrame
        if isinstance(fStartZ_or_frame, LocalFrame):
            oRefFrame = fStartZ_or_frame
            fStartZ   = float(fEndZ_or_startZ)
            fEndZ     = float(fRadius_or_endZ)
            fR        = float(fRadius) if fRadius is not None else 1.0
            lz        = as_np3(oRefFrame.vecGetLocalZ())
            oFrame    = oRefFrame.oTranslate(as_vec3(lz * fStartZ))
            length    = fEndZ - fStartZ
        else:
            fStartZ = float(fStartZ_or_frame)
            fEndZ   = float(fEndZ_or_startZ)
            fR      = float(fRadius_or_endZ)
            oFrame  = LocalFrame((0.0, 0.0, fStartZ))
            length  = fEndZ - fStartZ
        return BaseCylinder.from_frame(oFrame, length, fR).voxConstruct()

    @staticmethod
    def voxGetCone(fStartZ_or_frame: float | LocalFrame, fEndZ_or_startZ: float, fStartRadius_or_endZ: float, fEndRadius_or_startR: float, fEndRadius: float | None = None) -> Voxels:
        from ..base_shapes import BaseCone
        from ..frames.local_frames import LocalFrame
        if isinstance(fStartZ_or_frame, LocalFrame):
            oRefFrame = fStartZ_or_frame
            fStartZ   = float(fEndZ_or_startZ)
            fEndZ     = float(fStartRadius_or_endZ)
            fStartR   = float(fEndRadius_or_startR)
            fEndR     = float(fEndRadius) if fEndRadius is not None else fStartR
            lz        = as_np3(oRefFrame.vecGetLocalZ())
            oFrame    = oRefFrame.oTranslate(as_vec3(lz * fStartZ))
            length    = fEndZ - fStartZ
        else:
            fStartZ = float(fStartZ_or_frame)
            fEndZ   = float(fEndZ_or_startZ)
            fStartR = float(fStartRadius_or_endZ)
            fEndR   = float(fEndRadius_or_startR)
            oFrame  = LocalFrame((0.0, 0.0, fStartZ))
            length  = fEndZ - fStartZ
        return BaseCone(oFrame, length, fStartR, fEndR).voxConstruct()

    @staticmethod
    def voxGetPipe(fStartZ_or_frame: float | LocalFrame, fEndZ_or_startZ: float, fInnerRadius_or_endZ: float, fOuterRadius_or_innerR: float, fOuterRadius: float | None = None) -> Voxels:
        from ..base_shapes import BasePipe
        from ..frames.local_frames import LocalFrame
        if isinstance(fStartZ_or_frame, LocalFrame):
            oRefFrame = fStartZ_or_frame
            fStartZ   = float(fEndZ_or_startZ)
            fEndZ     = float(fInnerRadius_or_endZ)
            fInnerR   = float(fOuterRadius_or_innerR)
            fOuterR   = float(fOuterRadius) if fOuterRadius is not None else fInnerR
            lz        = as_np3(oRefFrame.vecGetLocalZ())
            oFrame    = oRefFrame.oTranslate(as_vec3(lz * fStartZ))
            length    = fEndZ - fStartZ
        else:
            fStartZ = float(fStartZ_or_frame)
            fEndZ   = float(fEndZ_or_startZ)
            fInnerR = float(fInnerRadius_or_endZ)
            fOuterR = float(fOuterRadius_or_innerR)
            oFrame  = LocalFrame((0.0, 0.0, fStartZ))
            length  = fEndZ - fStartZ
        return BasePipe.from_frame(oFrame, length, fInnerR, fOuterR).voxConstruct()




from __future__ import annotations

import math

from picogk import Mesh, Voxels

from .._types import Vec3, as_np3, as_vec3
from ..frames.local_frames import LocalFrame
from ..modulations import SurfaceModulation
from .base_shape import BaseShape


class BaseLens(BaseShape):
	def __init__(self, oFrame: LocalFrame, fHeight: float, fInnerRadius: float, fOuterRadius: float) -> None:
		super().__init__()
		self.m_oFrame = oFrame
		self.SetRadialSteps(5)
		self.SetPolarSteps(360)
		self.SetHeightSteps(5)
		self.m_fInnerRadius = fInnerRadius
		self.m_fOuterRadius = fOuterRadius
		self.m_oLowerModulation = SurfaceModulation(0.0)
		self.m_oUpperModulation = SurfaceModulation(fHeight)

	def SetHeight(self, oLowerModulation: SurfaceModulation, oUpperModulation: SurfaceModulation) -> None:
		self.m_oLowerModulation = oLowerModulation
		self.m_oUpperModulation = oUpperModulation
		self.SetRadialSteps(500)

	def SetRadialSteps(self, nRadialSteps: int) -> None:
		self.m_nRadialSteps = max(5, int(nRadialSteps))

	def SetPolarSteps(self, nPolarSteps: int) -> None:
		self.m_nPolarSteps = max(5, int(nPolarSteps))

	def SetHeightSteps(self, nHeightSteps: int) -> None:
		self.m_nHeightSteps = max(5, int(nHeightSteps))

	def voxConstruct(self) -> Voxels:
		return Voxels.from_mesh(self.mshConstruct())

	def mshConstruct(self) -> Mesh:
		mesh = Mesh()
		for ip in range(1, self.m_nPolarSteps):
			p1, p2 = self._phi_ratio(ip - 1), self._phi_ratio(ip)
			for ir in range(1, self.m_nRadialSteps):
				r1, r2 = self._radius_ratio(ir - 1), self._radius_ratio(ir)
				mesh.nAddTriangle(self.vecGetSurfacePoint(1.0, p1, r1), self.vecGetSurfacePoint(1.0, p1, r2), self.vecGetSurfacePoint(1.0, p2, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(1.0, p1, r1), self.vecGetSurfacePoint(1.0, p2, r2), self.vecGetSurfacePoint(1.0, p2, r1))
				mesh.nAddTriangle(self.vecGetSurfacePoint(0.0, p1, r1), self.vecGetSurfacePoint(0.0, p2, r2), self.vecGetSurfacePoint(0.0, p1, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(0.0, p1, r1), self.vecGetSurfacePoint(0.0, p2, r1), self.vecGetSurfacePoint(0.0, p2, r2))
			for ih in range(1, self.m_nHeightSteps):
				h1, h2 = self._height_ratio(ih - 1), self._height_ratio(ih)
				r_in = self._radius_ratio(0)
				r_out = self._radius_ratio(self.m_nRadialSteps - 1)
				mesh.nAddTriangle(self.vecGetSurfacePoint(h1, p1, r_in), self.vecGetSurfacePoint(h2, p1, r_in), self.vecGetSurfacePoint(h2, p2, r_in))
				mesh.nAddTriangle(self.vecGetSurfacePoint(h1, p1, r_in), self.vecGetSurfacePoint(h2, p2, r_in), self.vecGetSurfacePoint(h1, p2, r_in))
				mesh.nAddTriangle(self.vecGetSurfacePoint(h1, p1, r_out), self.vecGetSurfacePoint(h2, p2, r_out), self.vecGetSurfacePoint(h2, p1, r_out))
				mesh.nAddTriangle(self.vecGetSurfacePoint(h1, p1, r_out), self.vecGetSurfacePoint(h1, p2, r_out), self.vecGetSurfacePoint(h2, p2, r_out))
		return mesh

	def _radius_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nRadialSteps - 1)) * float(step)

	def _phi_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nPolarSteps - 1)) * float(step)

	def _height_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nHeightSteps - 1)) * float(step)

	def fGetHeight(self, fHeightRatio: float, fPhi: float, fRadiusRatio: float) -> float:
		low = self.m_oLowerModulation.fGetModulation(fPhi, fRadiusRatio)
		high = self.m_oUpperModulation.fGetModulation(fPhi, fRadiusRatio)
		return low + fHeightRatio * (high - low)

	def vecGetSurfacePoint(self, fHeightRatio: float, fPhiRatio: float, fRadiusRatio: float) -> Vec3:
		phi = 2.0 * math.pi * fPhiRatio
		radius = (self.m_fOuterRadius - self.m_fInnerRadius) * fRadiusRatio + self.m_fInnerRadius
		z = self.fGetHeight(fHeightRatio, phi, fRadiusRatio)
		x = radius * math.cos(phi)
		y = radius * math.sin(phi)
		pos = as_np3(self.m_oFrame.vecGetPosition())
		lx = as_np3(self.m_oFrame.vecGetLocalX())
		ly = as_np3(self.m_oFrame.vecGetLocalY())
		lz = as_np3(self.m_oFrame.vecGetLocalZ())
		return self.m_fnTrafo(as_vec3(pos + x * lx + y * ly + z * lz))


__all__ = ["BaseLens"]



from __future__ import annotations

import ctypes
from typing import Any, Callable
import numpy as np

from _common.types import Vector3Like

from .._base import HandleOwner
from .._errors import PicoGKInvalidHandleError
from .._native import CallbackScalarFieldTraverse, CallbackVectorFieldTraverse
from .._types import Vec3, VoxelDimensions, as_vec3, vec3_tuple

from .library import Library
from .metadata import FieldMetadata
from .._core.voxels import IImplicit, Voxels

class ScalarField(HandleOwner, IImplicit):
    def __init__(
        self,
        handle: int | ctypes.c_void_p | None = None,
        *,
        copy_from: "ScalarField | None" = None,
        from_voxels: Voxels | None = None,
        fill_value: float | None = None,
        sd_threshold: float = 0.5,
        owns_handle: bool = True,
    ) -> None:
        if handle is None and copy_from is not None:
            handle = Library._lib().ScalarField_hCreateCopy(copy_from.handle)
        elif handle is None and from_voxels is not None and fill_value is None:
            handle = Library._lib().ScalarField_hCreateFromVoxels(from_voxels.handle)
        elif handle is None and from_voxels is not None and fill_value is not None:
            handle = Library._lib().ScalarField_hBuildFromVoxels(from_voxels.handle, ctypes.c_float(fill_value), ctypes.c_float(sd_threshold))
        elif handle is None:
            handle = Library._lib().ScalarField_hCreate()

        if not handle:
            raise PicoGKInvalidHandleError("ScalarField handle creation returned null")
        super().__init__(handle, owns_handle=owns_handle)
        self.metadata = FieldMetadata.from_scalar_field(self)
        self.metadata.set_string("PicoGK.Class", "ScalarField")
        self._traverse_callback_ref: Any = None

    @classmethod
    def from_voxels(cls, voxels: Voxels) -> "ScalarField":
        return cls(from_voxels=voxels)

    @classmethod
    def build_from_voxels(cls, voxels: Voxels, value: float, sd_threshold: float = 0.5) -> "ScalarField":
        return cls(from_voxels=voxels, fill_value=value, sd_threshold=sd_threshold)

    def duplicate(self) -> "ScalarField":
        return ScalarField(copy_from=self)

    def _destroy_native(self) -> None:
        try:
            self.metadata.close()
        finally:
            Library._lib().ScalarField_Destroy(self._handle)

    def is_valid(self) -> bool:
        return bool(Library._lib().ScalarField_bIsValid(self.handle))

    def set_value(self, position: Vector3Like, value: float) -> "ScalarField":
        self._ensure_open()
        p = as_vec3(position)
        Library._lib().ScalarField_SetValue(self.handle, ctypes.byref(p), ctypes.c_float(value))
        return self

    SetValue = set_value

    def get_value(self, position: Vector3Like) -> tuple[bool, float]:
        self._ensure_open()
        p = as_vec3(position)
        out = ctypes.c_float(0.0)
        ok = bool(Library._lib().ScalarField_bGetValue(self.handle, ctypes.byref(p), ctypes.byref(out)))
        return ok, float(out.value)

    bGetValue = get_value

    def remove_value(self, position: Vector3Like) -> bool:
        self._ensure_open()
        p = as_vec3(position)
        return bool(Library._lib().ScalarField_RemoveValue(self.handle, ctypes.byref(p)))

    RemoveValue = remove_value

    def voxel_dimensions(self) -> VoxelDimensions:
        xo = ctypes.c_int()
        yo = ctypes.c_int()
        zo = ctypes.c_int()
        xs = ctypes.c_int()
        ys = ctypes.c_int()
        zs = ctypes.c_int()
        Library._lib().ScalarField_GetVoxelDimensions(self.handle, ctypes.byref(xo), ctypes.byref(yo), ctypes.byref(zo), ctypes.byref(xs), ctypes.byref(ys), ctypes.byref(zs))
        return VoxelDimensions(xo.value, yo.value, zo.value, xs.value, ys.value, zs.value)

    GetVoxelDimensions = voxel_dimensions

    def slice(self, z_slice: int) -> np.ndarray:
        dims = self.voxel_dimensions()
        count = dims.x_size * dims.y_size
        buf = (ctypes.c_float * count)()
        Library._lib().ScalarField_GetSlice(self.handle, int(z_slice), ctypes.cast(buf, ctypes.c_void_p))
        return np.frombuffer(buf, dtype=np.float32).reshape((dims.y_size, dims.x_size)).copy()

    GetVoxelSlice = slice

    def traverse_active(self, callback: Callable[[tuple[float, float, float], float], None]) -> "ScalarField":
        def _wrapped(vec_ptr: Any, value: float) -> None:
            vec = vec_ptr.contents
            callback((float(vec.x), float(vec.y), float(vec.z)), float(value))
        self._traverse_callback_ref = CallbackScalarFieldTraverse(_wrapped)
        Library._lib().ScalarField_TraverseActive(self.handle, self._traverse_callback_ref)
        return self

    TraverseActive = traverse_active

    def signed_distance(self, vecPt: Vector3Like) -> float:
        ok, value = self.get_value(vecPt)
        return float(value if ok else 0.0) * Library.voxel_size_mm

    fSignedDistance = signed_distance

    def bounding_box(self) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        dims = self.voxel_dimensions()
        return (
            Library.voxels_to_mm((dims.x_origin, dims.y_origin, dims.z_origin)),
            Library.voxels_to_mm((dims.x_origin + dims.x_size, dims.y_origin + dims.y_size, dims.z_origin + dims.z_size)),
        )

    oBoundingBox = bounding_box

    def oMetaData(self) -> FieldMetadata:
        return self.metadata


class VectorField(HandleOwner):
    def __init__(
        self,
        handle: int | ctypes.c_void_p | None = None,
        *,
        copy_from: "VectorField | None" = None,
        from_voxels: Voxels | None = None,
        fill_value: Vector3Like | None = None,
        sd_threshold: float = 0.5,
        owns_handle: bool = True,
    ) -> None:
        if handle is None and copy_from is not None:
            handle = Library._lib().VectorField_hCreateCopy(copy_from.handle)
        elif handle is None and from_voxels is not None and fill_value is None:
            handle = Library._lib().VectorField_hCreateFromVoxels(from_voxels.handle)
        elif handle is None and from_voxels is not None and fill_value is not None:
            fill = as_vec3(fill_value)
            handle = Library._lib().VectorField_hBuildFromVoxels(from_voxels.handle, ctypes.byref(fill), ctypes.c_float(sd_threshold))
        elif handle is None:
            handle = Library._lib().VectorField_hCreate()

        if not handle:
            raise PicoGKInvalidHandleError("VectorField handle creation returned null")
        super().__init__(handle, owns_handle=owns_handle)
        self.metadata = FieldMetadata.from_vector_field(self)
        self.metadata.set_string("PicoGK.Class", "VectorField")
        self._traverse_callback_ref: Any = None

    @classmethod
    def from_voxels(cls, voxels: Voxels) -> "VectorField":
        return cls(from_voxels=voxels)

    @classmethod
    def build_from_voxels(cls, voxels: Voxels, value: Vector3Like, sd_threshold: float = 0.5) -> "VectorField":
        return cls(from_voxels=voxels, fill_value=value, sd_threshold=sd_threshold)

    def duplicate(self) -> "VectorField":
        return VectorField(copy_from=self)

    def _destroy_native(self) -> None:
        try:
            self.metadata.close()
        finally:
            Library._lib().VectorField_Destroy(self._handle)

    def is_valid(self) -> bool:
        return bool(Library._lib().VectorField_bIsValid(self.handle))

    def set_value(self, position: Vector3Like, value: Vector3Like) -> "VectorField":
        self._ensure_open()
        p = as_vec3(position)
        v = as_vec3(value)
        Library._lib().VectorField_SetValue(self.handle, ctypes.byref(p), ctypes.byref(v))
        return self

    SetValue = set_value

    def get_value(self, position: Vector3Like) -> tuple[bool, tuple[float, float, float]]:
        self._ensure_open()
        p = as_vec3(position)
        out = Vec3()
        ok = bool(Library._lib().VectorField_bGetValue(self.handle, ctypes.byref(p), ctypes.byref(out)))
        return ok, vec3_tuple(out)

    bGetValue = get_value

    def remove_value(self, position: Vector3Like) -> bool:
        self._ensure_open()
        p = as_vec3(position)
        return bool(Library._lib().VectorField_RemoveValue(self.handle, ctypes.byref(p)))

    RemoveValue = remove_value

    def traverse_active(self, callback: Callable[[tuple[float, float, float], tuple[float, float, float]], None]) -> "VectorField":
        def _wrapped(pos_ptr: Any, val_ptr: Any) -> None:
            pos = pos_ptr.contents
            val = val_ptr.contents
            callback((float(pos.x), float(pos.y), float(pos.z)), (float(val.x), float(val.y), float(val.z)))
        self._traverse_callback_ref = CallbackVectorFieldTraverse(_wrapped)
        Library._lib().VectorField_TraverseActive(self.handle, self._traverse_callback_ref)
        return self

    TraverseActive = traverse_active

    def oMetaData(self) -> FieldMetadata:
        return self.metadata






from __future__ import annotations

from abc import ABC, abstractmethod

from .._types import Vec3


class ISpectrum(ABC):
	@abstractmethod
	def aGetRawRGBList(self) -> list[Vec3]:
		raise NotImplementedError


class RainboxSpectrum(ISpectrum):
	def aGetRawRGBList(self) -> list[Vec3]:
		return [
			(0.0, 0.0, 255.0),
			(0.0, 255.0, 0.0),
			(255.0, 255.0, 0.0),
			(255.0, 130.0, 0.0),
			(255.0, 0.0, 0.0),
		]


__all__ = ["ISpectrum", "RainboxSpectrum"]



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

import numpy as np
from ..splines import ControlPointSpline

from .._types import Vec3, Vector3Like, as_np3, as_vec3

if TYPE_CHECKING:
    from ..frames.local_frames import LocalFrame
    
from .vec_operations import VecOperations

from picogk import Voxels

class SplineOperations:
    @staticmethod
    def aGetReparametrizedSpline(aPoints: Sequence[Vector3Like], target: int | float) -> list[Vec3]:
        points = [as_np3(p) for p in aPoints]
        if len(points) < 2:
            return [as_vec3(p) for p in points]

        if type(target) is float:
            total_length = SplineOperations.fGetTotalLength(aPoints)
            n_target_samples = int(max(10.0, total_length / max(float(target), 1e-6)))
        else:
            n_target_samples = int(target)

        lengths = SplineOperations.aGetLengthsAtIndices(aPoints)
        spine_len = lengths[-1]
        target_step = spine_len / max(1, n_target_samples)

        out: list[np.ndarray] = [points[0]]
        for sample_idx in range(1, n_target_samples):
            target_len = target_step * sample_idx
            upper_idx = 1
            while upper_idx < len(points) and lengths[upper_idx] < target_len:
                upper_idx += 1
            upper_idx = min(upper_idx, len(points) - 1)
            lower_idx = max(upper_idx - 1, 0)

            lower_len = lengths[lower_idx]
            upper_len = lengths[upper_idx]
            if abs(upper_len - lower_len) <= 1e-12:
                ds = 0.0
            else:
                ds = (target_len - lower_len) / (upper_len - lower_len)
            out.append(points[lower_idx] + ds * (points[upper_idx] - points[lower_idx]))

        out.append(points[-1])
        return [as_vec3(p) for p in out]

    @staticmethod
    def aGetNURBSpline(aControlPoints: Sequence[Vector3Like], nSamples: int) -> list[Vec3]:
        return ControlPointSpline(aControlPoints, 2, ControlPointSpline.EEnds.OPEN).aGetPoints(nSamples)

    @staticmethod
    def aGetLengthsAtIndices(aPoints: Sequence[Vector3Like]) -> list[float]:
        points = [as_np3(p) for p in aPoints]
        out = [0.0]
        total = 0.0
        for idx in range(1, len(points)):
            total += float(np.linalg.norm(points[idx] - points[idx - 1]))
            out.append(total)
        return out

    @staticmethod
    def fGetTotalLength(aPoints: Sequence[Vector3Like]) -> float:
        lengths = SplineOperations.aGetLengthsAtIndices(aPoints)
        return lengths[-1] if lengths else 0.0

    @staticmethod
    def aGetLinearInterpolation(vecStart: Vector3Like, vecEnd: Vector3Like, nSamples: int) -> list[Vec3]:
        start = as_np3(vecStart)
        end   = as_np3(vecEnd)
        n     = max(2, int(nSamples))
        return [as_vec3(start + (float(i) / float(n - 1)) * (end - start)) for i in range(n)]

    @staticmethod
    def aGetSnappedSpline(aPoints: Sequence[Vector3Like], voxTarget: "Voxels") -> list[Vec3]:
        out: list[Vec3] = []
        for pt in aPoints:
            p = as_np3(pt)
            _ok, snap = voxTarget.closest_point_on_surface((float(p[0]), float(p[1]), float(p[2])))
            out.append(snap)
        return out

    @staticmethod
    def aGetAveragePointSpacing(aPoints: Sequence[Vector3Like]) -> float:
        pts = list(aPoints)
        if len(pts) < 2:
            return 0.0
        return SplineOperations.fGetTotalLength(pts) / float(len(pts) - 1)

    @staticmethod
    def aSplitLists(aList: Sequence[Vector3Like], iFirstIndexOfSecondList: int) -> tuple[list[Vec3], list[Vec3]]:
        idx    = int(iFirstIndexOfSecondList)
        first  = [as_vec3(aList[i]) for i in range(len(aList)) if i < idx]
        second = [as_vec3(aList[i]) for i in range(len(aList)) if i >= idx]
        return first, second

    @staticmethod
    def aCombineLists(aLists: Sequence[Sequence[Vector3Like]]) -> list[Vec3]:
        out: list[Vec3] = []
        for lst in aLists:
            for pt in lst:
                out.append(as_vec3(pt))
        return out

    @staticmethod
    def aRotateListAroundZ(aList: Sequence[Vector3Like], fAngle: float) -> list[Vec3]:
        return [VecOperations.vecRotateAroundZ(pt, float(fAngle)) for pt in aList]

    @staticmethod
    def aTranslateList(aList: Sequence[Vector3Like], vecShift: Vector3Like) -> list[Vec3]:
        shift = as_np3(vecShift)
        return [as_vec3(as_np3(pt) + shift) for pt in aList]

    @staticmethod
    def aScaleList(aList: Sequence[Vector3Like], fScale: float) -> list[Vec3]:
        s = float(fScale)
        return [as_vec3(as_np3(pt) * s) for pt in aList]

    @staticmethod
    def aOverSampleList(aList: Sequence[Vector3Like], iSamplesPerStep: int) -> list[Vec3]:
        out: list[Vec3] = []
        pts = list(aList)
        n   = int(iSamplesPerStep)
        for i in range(1, len(pts)):
            for j in range(n):
                t = float(j) / float(n)
                out.append(as_vec3(as_np3(pts[i - 1]) + t * (as_np3(pts[i]) - as_np3(pts[i - 1]))))
        out.append(as_vec3(pts[-1]))
        return out

    @staticmethod
    def aSubSampleList(aList: Sequence[Vector3Like], iSampleSize: int) -> list[Vec3]:
        pts  = list(aList)
        step = max(1, int(iSampleSize))
        out  = [as_vec3(pts[i]) for i in range(0, len(pts), step)]
        if tuple(as_np3(out[-1])) != tuple(as_np3(pts[-1])):
            out.append(as_vec3(pts[-1]))
        return out

    @staticmethod
    def aTranslateListOntoFrame(oFrame: LocalFrame, aList: Sequence[Vector3Like]) -> list[Vec3]:
        return [VecOperations.vecTranslatePointOntoFrame(oFrame, pt) for pt in aList]

    @staticmethod
    def aExpressListInFrame(oFrame: LocalFrame, aList: Sequence[Vector3Like]) -> list[Vec3]:
        return [VecOperations.vecExpressPointInFrame(oFrame, pt) for pt in aList]

    @staticmethod
    def aRotateListAroundAxis(
        aList: Sequence[Vector3Like],
        dPhi: float,
        vecAxis: Vector3Like,
        vecAxisOrigin: Vector3Like | None = None,
    ) -> list[Vec3]:
        return [VecOperations.vecRotateAroundAxis(pt, float(dPhi), vecAxis, vecAxisOrigin) for pt in aList]

    @staticmethod
    def vecGetAverage(aPoints: Sequence[Vector3Like]) -> Vec3:
        pts = [as_np3(p) for p in aPoints]
        if not pts:
            return (0.0, 0.0, 0.0)
        return as_vec3(np.mean(np.stack(pts, axis=0), axis=0))

    @staticmethod
    def vecGetClosestPoint(aPoints: Sequence[Vector3Like], vecStart: Vector3Like) -> Vec3:
        start    = as_np3(vecStart)
        best_d   = float("inf")
        best_pt  = as_np3(aPoints[0])
        for pt in aPoints:
            p = as_np3(pt)
            d = float(np.dot(p - start, p - start))
            if d < best_d:
                best_d  = d
                best_pt = p
        return as_vec3(best_pt)

    @staticmethod
    def fGetDistanceToClosestPoint(aPoints: Sequence[Vector3Like], vecStart: Vector3Like) -> float:
        start  = as_np3(vecStart)
        min_sq = float("inf")
        for pt in aPoints:
            p = as_np3(pt)
            d = float(np.dot(p - start, p - start))
            if d < min_sq:
                min_sq = d
        import math as _math
        return float(_math.sqrt(min_sq))

    @staticmethod
    def aGetClusteredPoints(aPoints: Sequence[Vector3Like], fClusteringRange: float) -> list[Vec3]:
        threshold = float(fClusteringRange) ** 2
        result    = [as_vec3(aPoints[0])]
        for pt in aPoints[1:]:
            near = as_np3(SplineOperations.vecGetClosestPoint(result, pt))
            diff = as_np3(pt) - near
            if float(np.dot(diff, diff)) > threshold:
                result.append(as_vec3(pt))
        return result



from __future__ import annotations

import math

from picogk import Mesh, Voxels

from .._types import Vec3, as_np3, as_vec3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from ..modulations import SurfaceModulation
from .base_shape import BaseShape


class BaseCylinder(BaseShape):
	def __init__(
		self,
		aFrames: Frames,
		fRadius: float,
		*,
		_length_steps: int,
	) -> None:
		super().__init__()
		self.m_aFrames = aFrames
		self.SetLengthSteps(_length_steps)
		self.SetPolarSteps(360)
		self.SetRadialSteps(5)
		self.m_oRadiusModulation = SurfaceModulation(fRadius)

	@classmethod
	def from_frame(
		cls,
		oFrame: LocalFrame,
		fLength: float = 20.0,
		fRadius: float = 10.0,
	) -> "BaseCylinder":
		return cls(Frames.from_length(fLength, oFrame), fRadius, _length_steps=5)

	@classmethod
	def from_frames(
		cls,
		aFrames: Frames,
		fRadius: float = 10.0,
	) -> "BaseCylinder":
		return cls(aFrames, fRadius, _length_steps=500)

	def SetRadius(self, oModulation: SurfaceModulation) -> None:
		self.m_oRadiusModulation = oModulation
		self.SetLengthSteps(500)

	def SetRadialSteps(self, nRadialSteps: int) -> None:
		self.m_nRadialSteps = max(5, int(nRadialSteps))

	def SetPolarSteps(self, nPolarSteps: int) -> None:
		self.m_nPolarSteps = max(5, int(nPolarSteps))

	def SetLengthSteps(self, nLengthSteps: int) -> None:
		self.m_nLengthSteps = max(5, int(nLengthSteps))

	def voxConstruct(self) -> Voxels:
		return Voxels.from_mesh(self.mshConstruct())

	def mshConstruct(self) -> Mesh:
		mesh = Mesh()
		self._add_top_surface(mesh)
		self._add_outer_mantle(mesh)
		self._add_bottom_surface(mesh)
		return mesh

	def _add_top_surface(self, mesh: Mesh) -> None:
		lr = self._length_ratio(self.m_nLengthSteps - 1)
		for ip in range(1, self.m_nPolarSteps):
			p1, p2 = self._phi_ratio(ip - 1), self._phi_ratio(ip)
			for ir in range(1, self.m_nRadialSteps):
				r1, r2 = self._radius_ratio(ir - 1), self._radius_ratio(ir)
				mesh.nAddTriangle(self.vecGetSurfacePoint(lr, p1, r1), self.vecGetSurfacePoint(lr, p1, r2), self.vecGetSurfacePoint(lr, p2, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(lr, p1, r1), self.vecGetSurfacePoint(lr, p2, r2), self.vecGetSurfacePoint(lr, p2, r1))

	def _add_bottom_surface(self, mesh: Mesh) -> None:
		lr = self._length_ratio(0)
		for ip in range(1, self.m_nPolarSteps):
			p1, p2 = self._phi_ratio(ip - 1), self._phi_ratio(ip)
			for ir in range(1, self.m_nRadialSteps):
				r1, r2 = self._radius_ratio(ir - 1), self._radius_ratio(ir)
				mesh.nAddTriangle(self.vecGetSurfacePoint(lr, p1, r1), self.vecGetSurfacePoint(lr, p2, r2), self.vecGetSurfacePoint(lr, p1, r2))
				mesh.nAddTriangle(self.vecGetSurfacePoint(lr, p1, r1), self.vecGetSurfacePoint(lr, p2, r1), self.vecGetSurfacePoint(lr, p2, r2))

	def _add_outer_mantle(self, mesh: Mesh) -> None:
		rr = self._radius_ratio(self.m_nRadialSteps - 1)
		for ip in range(1, self.m_nPolarSteps):
			p1, p2 = self._phi_ratio(ip - 1), self._phi_ratio(ip)
			for il in range(1, self.m_nLengthSteps):
				l1, l2 = self._length_ratio(il - 1), self._length_ratio(il)
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, p1, rr), self.vecGetSurfacePoint(l2, p2, rr), self.vecGetSurfacePoint(l2, p1, rr))
				mesh.nAddTriangle(self.vecGetSurfacePoint(l1, p1, rr), self.vecGetSurfacePoint(l1, p2, rr), self.vecGetSurfacePoint(l2, p2, rr))

	def _radius_ratio(self, step: int) -> float:
		return (1.0 / float(self.m_nRadialSteps - 1)) * float(step)

	def _phi_ratio(self, step: int) -> float:
		return (1.0 / float(self.m_nPolarSteps - 1)) * float(step)

	def _length_ratio(self, step: int) -> float:
		return (1.0 / float(self.m_nLengthSteps - 1)) * float(step)

	def fGetRadius(self, fPhi: float, fLengthRatio: float) -> float:
		return self.m_oRadiusModulation.fGetModulation(fPhi, fLengthRatio)

	def vecGetSurfacePoint(self, fLengthRatio: float, fPhiRatio: float, fRadiusRatio: float) -> Vec3:
		phi = 2.0 * math.pi * fPhiRatio
		spine = as_np3(self.m_aFrames.vecGetSpineAlongLength(fLengthRatio))
		lx = as_np3(self.m_aFrames.vecGetLocalXAlongLength(fLengthRatio))
		ly = as_np3(self.m_aFrames.vecGetLocalYAlongLength(fLengthRatio))
		radius = fRadiusRatio * self.fGetRadius(phi, fLengthRatio)
		pt = spine + (radius * math.cos(phi)) * lx + (radius * math.sin(phi)) * ly
		return self.m_fnTrafo(as_vec3(pt))


__all__ = ["BaseCylinder"]



from __future__ import annotations


class CSVWriter:
	def __init__(self, strFilename: str) -> None:
		self._filename = str(strFilename)
		self._writer = open(self._filename, "w", newline="", encoding="utf-8")

	def AddLine(self, strLine: str) -> None:
		self._writer.write(str(strLine) + "\n")

	def ExportCSVFile(self) -> None:
		try:
			self._writer.flush()
		except Exception as exc:
			print(f"Could not save CSV: {exc}")

	def __del__(self) -> None:
		try:
			self._writer.close()
		except Exception:
			pass


__all__ = ["CSVWriter"]



from __future__ import annotations

import numpy as np
from typing import TYPE_CHECKING

from picogk import Library, Mesh, Voxels

from .._types import Vec3, Vector3Like, as_np3, as_vec3

if TYPE_CHECKING:
    from ..frames.local_frames import LocalFrame
    
from .vec_operations import VecOperations


class Measure:
    @staticmethod
    def fGetVolume(oVoxels: Voxels) -> float:
        volume, _bbox = oVoxels.calculate_properties()
        return float(volume)

    @staticmethod
    def fGetSurfaceArea(oVoxels_or_mesh: Voxels | Mesh) -> float:
        if isinstance(oVoxels_or_mesh, Voxels):
            with Mesh.from_voxels(oVoxels_or_mesh) as msh:
                return Measure.fGetSurfaceArea(msh)
        msh = oVoxels_or_mesh
        total = 0.0
        for i in range(msh.triangle_count()):
            a, b, c = msh.get_triangle_vertices(i)
            total += Measure.fGetTriangleArea(a, b, c)
        return total

    @staticmethod
    def fGetTriangleArea(vecA: Vector3Like, vecB: Vector3Like, vecC: Vector3Like) -> float:
        ab = as_np3(vecB) - as_np3(vecA)
        ac = as_np3(vecC) - as_np3(vecA)
        return float(0.5 * np.linalg.norm(np.cross(ab, ac)))

    @staticmethod
    def vecGetCentreOfGravity(oVoxels: Voxels) -> Vec3:
        _volume, (box_min, box_max) = oVoxels.calculate_properties()
        step = Library.voxel_size_mm
        bmin = np.array(box_min, dtype=np.float64) - step
        bmax = np.array(box_max, dtype=np.float64) + step
        cog  = np.zeros(3, dtype=np.float64)
        count = 0
        xs = np.arange(float(bmin[0]), float(bmax[0]), step)
        ys = np.arange(float(bmin[1]), float(bmax[1]), step)
        zs = np.arange(float(bmin[2]), float(bmax[2]), step)
        for x in xs:
            for y in ys:
                for z in zs:
                    pt = (float(x), float(y), float(z))
                    if oVoxels.is_inside(pt):
                        cog += np.array([x, y, z])
                        count += 1
        if count > 0:
            cog /= count
        return as_vec3(cog)

    @staticmethod
    def matGetMomentOfInertia(oVoxels: Voxels, oRefFrame: LocalFrame, fDensity: float) -> list[list[float]]:
        volume, (box_min, box_max) = oVoxels.calculate_properties()
        step       = Library.voxel_size_mm
        bmin       = np.array(box_min, dtype=np.float64) - step
        bmax       = np.array(box_max, dtype=np.float64) + step
        dv         = step ** 3
        mat        = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
        vox_count  = 0
        xs = np.arange(float(bmin[0]), float(bmax[0]), step)
        ys = np.arange(float(bmin[1]), float(bmax[1]), step)
        zs = np.arange(float(bmin[2]), float(bmax[2]), step)
        for x in xs:
            for y in ys:
                for z in zs:
                    pt = (float(x), float(y), float(z))
                    if oVoxels.is_inside(pt):
                        vox_count += 1
        vox_vol    = float(vox_count) * dv
        vol_factor = float(volume) / max(vox_vol, 1e-30)
        for x in xs:
            for y in ys:
                for z in zs:
                    pt = (float(x), float(y), float(z))
                    if oVoxels.is_inside(pt):
                        dm  = float(fDensity) * vol_factor * dv / 1e9
                        rel = VecOperations.vecExpressPointInFrame(oRefFrame, pt)
                        rx, ry, rz = float(rel[0]), float(rel[1]), float(rel[2])
                        mat[0][0] += dm * (ry * ry + rz * rz)
                        mat[1][1] += dm * (rx * rx + rz * rz)
                        mat[2][2] += dm * (rx * rx + ry * ry)
                        mat[0][1] += dm * (-rx * ry)
                        mat[0][2] += dm * (-rx * rz)
                        mat[1][0] += dm * (-ry * rx)
                        mat[1][2] += dm * (-ry * rz)
                        mat[2][0] += dm * (-rz * rx)
                        mat[2][1] += dm * (-rz * ry)
        return mat





import numpy as np

from _common.types import Vector3Like

from .mesh import Mesh
from .voxels import IBoundedImplicit, Voxels


class ImplicitTriangle(IBoundedImplicit):
    def __init__(self, vecA: Vector3Like, vecB: Vector3Like, vecC: Vector3Like, fThickness: float) -> None:
        self._a = np.array(vecA, dtype=np.float64)
        self._b = np.array(vecB, dtype=np.float64)
        self._c = np.array(vecC, dtype=np.float64)
        self._thickness = float(fThickness)

        bounds_min = np.minimum(np.minimum(self._a, self._b), self._c) - self._thickness
        bounds_max = np.maximum(np.maximum(self._a, self._b), self._c) + self._thickness
        self._bounds = (tuple(bounds_min.tolist()), tuple(bounds_max.tolist()))

    @property
    def oBounds(self) -> tuple[Vector3Like, Vector3Like]:
        return self._bounds

    @staticmethod
    def _closest_point_on_triangle(point: np.ndarray, a: np.ndarray, b: np.ndarray, c: np.ndarray) -> np.ndarray:
        ab = b - a
        ac = c - a
        ap = point - a
        d1 = np.dot(ab, ap)
        d2 = np.dot(ac, ap)
        if d1 <= 0.0 and d2 <= 0.0:
            return a
        bp = point - b
        d3 = np.dot(ab, bp)
        d4 = np.dot(ac, bp)
        if d3 >= 0.0 and d4 <= d3:
            return b
        vc = d1 * d4 - d3 * d2
        if vc <= 0.0 and d1 >= 0.0 and d3 <= 0.0:
            v = d1 / (d1 - d3)
            return a + (v * ab)
        cp = point - c
        d5 = np.dot(ab, cp)
        d6 = np.dot(ac, cp)
        if d6 >= 0.0 and d5 <= d6:
            return c
        vb = d5 * d2 - d1 * d6
        if vb <= 0.0 and d2 >= 0.0 and d6 <= 0.0:
            w = d2 / (d2 - d6)
            return a + (w * ac)
        va = d3 * d6 - d5 * d4
        if va <= 0.0 and (d4 - d3) >= 0.0 and (d5 - d6) >= 0.0:
            w = (d4 - d3) / ((d4 - d3) + (d5 - d6))
            return b + (w * (c - b))
        denom = 1.0 / (va + vb + vc)
        v_ab = vb * denom
        v_ac = vc * denom
        return a + (v_ab * ab) + (v_ac * ac)

    def fSignedDistance(self, vec: Vector3Like) -> float:
        p = np.array(vec, dtype=np.float64)
        cp = self._closest_point_on_triangle(p, self._a, self._b, self._c)
        return float(np.linalg.norm(p - cp) - self._thickness)


class ImplicitMesh(IBoundedImplicit):
    def __init__(self, msh: Mesh, fThickness: float) -> None:
        self._triangles: list[ImplicitTriangle] = []
        bounds_min = np.array((float("inf"), float("inf"), float("inf")), dtype=np.float64)
        bounds_max = np.array((float("-inf"), float("-inf"), float("-inf")), dtype=np.float64)

        for i in range(msh.nTriangleCount()):
            a, b, c = msh.GetTriangle(i)
            tri = ImplicitTriangle(a, b, c, fThickness)
            self._triangles.append(tri)
            tmin = np.array(tri.oBounds[0], dtype=np.float64)
            tmax = np.array(tri.oBounds[1], dtype=np.float64)
            bounds_min = np.minimum(bounds_min, tmin)
            bounds_max = np.maximum(bounds_max, tmax)

        self._bounds = (tuple(bounds_min.tolist()), tuple(bounds_max.tolist()))

    @property
    def oBounds(self) -> tuple[Vector3Like, Vector3Like]:
        return self._bounds

    def fSignedDistance(self, vec: Vector3Like) -> float:
        return min((tri.fSignedDistance(vec) for tri in self._triangles), default=1e9)


class TriangleVoxelization:
    @staticmethod
    def voxVoxelizeHollow(mesh: Mesh, fThickness: float) -> Voxels:
        implicit = ImplicitMesh(mesh, fThickness)
        return Voxels.from_bounded_implicit(implicit)





from .bisection import Bisection, BisectionException
from .csv_writer import CSVWriter
from .grid_operations import GridOperations
from .implicit_utility import ImplicitGenus, ImplicitGyroid, ImplicitSphere, ImplicitSuperEllipsoid
from .line_decimation import LineDecimation
from .list_operations import ListOperations
from .useful_formulas import Uf
from .vec_operations import VecOperations

from .cyl_utility import CylUtility
from .measure import Measure
from .mesh_utility import MeshUtility
from .spline_operations import SplineOperations

__all__ = [
    "Bisection",
    "BisectionException",
    "CSVWriter",
    "GridOperations",
    "ImplicitGenus",
    "ImplicitGyroid",
    "ImplicitSphere",
    "ImplicitSuperEllipsoid",
    "LineDecimation",
    "ListOperations",
    "Uf",
    "VecOperations",
    "CylUtility",
    "Measure",
    "MeshUtility",
    "SplineOperations",
]



from __future__ import annotations

import math

import numpy as np
from picogk import Mesh, Voxels

from .._types import Vec3, as_np3, as_vec3
from ..frames.local_frames import LocalFrame
from ..modulations import SurfaceModulation
from .base_shape import BaseShape


class BaseRing(BaseShape):
	def __init__(self, oFrame: LocalFrame, fRingRadius: float = 50.0, fRadius: float = 5.0) -> None:
		super().__init__()
		self.SetRadialSteps(360)
		self.SetPolarSteps(360)
		self.m_oFrame = oFrame
		self.m_fRingRadius = fRingRadius
		self.m_oRadiusModulation = SurfaceModulation(fRadius)

	def SetRadius(self, oModulation: SurfaceModulation) -> None:
		self.m_oRadiusModulation = oModulation

	def SetRadialSteps(self, nRadialSteps: int) -> None:
		self.m_nRadialSteps = max(5, int(nRadialSteps))

	def SetPolarSteps(self, nPolarSteps: int) -> None:
		self.m_nPolarSteps = max(5, int(nPolarSteps))

	def voxConstruct(self) -> Voxels:
		return Voxels.from_mesh(self.mshConstruct())

	def mshConstruct(self) -> Mesh:
		mesh = Mesh()
		rr = 1.0
		for ia in range(self.m_nRadialSteps):
			lower = ia - 1
			if lower < 0:
				lower += self.m_nRadialSteps
			a1 = self._alpha_ratio(lower)
			a2 = self._alpha_ratio(ia)
			for ip in range(1, self.m_nPolarSteps):
				p1, p2 = self._phi_ratio(ip - 1), self._phi_ratio(ip)
				mesh.nAddTriangle(self.vecGetSurfacePoint(a1, p1, rr), self.vecGetSurfacePoint(a2, p1, rr), self.vecGetSurfacePoint(a2, p2, rr))
				mesh.nAddTriangle(self.vecGetSurfacePoint(a1, p1, rr), self.vecGetSurfacePoint(a2, p2, rr), self.vecGetSurfacePoint(a1, p2, rr))
		return mesh

	def _alpha_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nRadialSteps - 1)) * float(step)

	def _phi_ratio(self, step: int) -> float:
		return (1.0 / (self.m_nPolarSteps - 1)) * float(step)

	def fGetRadius(self, fPhi: float, fLengthRatio: float) -> float:
		return self.m_oRadiusModulation.fGetModulation(fPhi, fLengthRatio)

	def vecGetSurfacePoint(self, fAlphaRatio: float, fPhiRatio: float, fRadiusRatio: float) -> Vec3:
		alpha = 2.0 * math.pi * fAlphaRatio
		phi = 2.0 * math.pi * fPhiRatio
		x = self.m_fRingRadius * math.cos(alpha)
		y = self.m_fRingRadius * math.sin(alpha)
		frame_pos = as_np3(self.m_oFrame.vecGetPosition())
		frame_x = as_np3(self.m_oFrame.vecGetLocalX())
		frame_y = as_np3(self.m_oFrame.vecGetLocalY())
		frame_z = as_np3(self.m_oFrame.vecGetLocalZ())
		spine = frame_pos + x * frame_x + y * frame_y
		local_x = spine - frame_pos
		norm = float(np.linalg.norm(local_x))
		if norm > 1e-12:
			local_x = local_x / norm
		local_y = frame_z
		radius = fRadiusRatio * self.fGetRadius(phi, alpha)
		pt = spine + (radius * math.cos(phi)) * local_x + (radius * math.sin(phi)) * local_y
		return self.m_fnTrafo(as_vec3(pt))


__all__ = ["BaseRing"]



from .animation import Animation, AnimationQueue
from .config import Config
from .bbox import BBox2, BBox3
from .cli import CliIo, Cli
from .csv import CsvTable
from .easing import Easing
from .field_utils import FieldUtils, SdfVisualizer, AddVectorFieldToViewer, ActiveVoxelCounterScalar, VectorFieldMerge, SurfaceNormalFieldExtractor
from .image import Image, ImageBW, ImageColor,ImageGrayScale,ImageRgb24,ImageRgba32
from .image_io import TgaIo
from .log import LogFile
from .mesh_io import MeshIo, EStlUnit
from .mesh_mat import MeshMath
from .slice import PolyContour, PolySlice, PolySliceStack
from .triangle_voxelization import TriangleVoxelization, ImplicitMesh, ImplicitTriangle
from .utils import Utils, TempFolder
from .vector3_ext import Vector3Ext
from .voxels import Voxels, ESliceMode, IBoundedImplicit, IImplicit
from .library import Library
from .metadata import FieldMetadata
from .mesh import Mesh
from .lattice import Lattice
from .fields import ScalarField, VectorField
from .openvdb import OpenVdbFile
from .polyline import PolyLine

__all__ = [
    "IImplicit",
    "IBoundedImplicit",
    "Library",
    "FieldMetadata",
    "Mesh",
    "Lattice",
    "Voxels",
    "ScalarField",
    "VectorField",
    "OpenVdbFile",
    "PolyLine",
    "Animation",
    "AnimationQueue",
    "Config",
    "BBox2",
    "BBox3",
    "Voxels",
    "ESliceMode",
    "CliIo",
    "Cli",
    "CsvTable",
    "Easing",
    "SdfVisualizer",
    "ActiveVoxelCounterScalar",
    "AddVectorFieldToViewer",
    "VectorFieldMerge",
    "SurfaceNormalFieldExtractor",
    "FieldUtils",
    "Image",
    "ImageBW",
    "ImageColor",
    "ImageGrayScale",
    "ImageRgb24",
    "ImageRgba32",
    "MeshIo",
    "EStlUnit",
    "MeshMath",
    "PolyContour",
    "PolySlice",
    "PolySliceStack",
    "TgaIo",
    "LogFile",
    "TempFolder",
    "Utils",
    "Vector3Ext",
    "TriangleVoxelization",
    "ImplicitMesh",
    "ImplicitTriangle"
]


from __future__ import annotations

from abc import ABC, abstractmethod

from shape_kernel._types import Vector3Like, as_vec3
from shape_kernel.utils import Uf, VecOperations


class ICoordinateTrafo(ABC):
	@abstractmethod
	def Apply(self, vecPt: Vector3Like) -> tuple[float, float, float]:
		raise NotImplementedError


class ScaleTrafo(ICoordinateTrafo):
	def __init__(self, fUnitX: float, fUnitY: float, fUnitZ: float) -> None:
		self.m_fUnitX = float(fUnitX)
		self.m_fUnitY = float(fUnitY)
		self.m_fUnitZ = float(fUnitZ)

	def Apply(self, vecPt: Vector3Like) -> tuple[float, float, float]:
		pt = as_vec3(vecPt)
		return (
			float(pt[0] / self.m_fUnitX),
			float(pt[1] / self.m_fUnitY),
			float(pt[2] / self.m_fUnitZ),
		)


class FunctionalScaleTrafo(ICoordinateTrafo):
	def Apply(self, vecPt: Vector3Like) -> tuple[float, float, float]:
		pt = as_vec3(vecPt)
		fRatio = min(max(float(pt[2]) / 50.0, 0.0), 1.0)
		fUnit = Uf.fTransFixed(20.0, 5.0, fRatio)
		return (float(pt[0] / fUnit), float(pt[1] / fUnit), 10.0)


class RadialTrafo(ICoordinateTrafo):
	def __init__(self, nSamplesPerRound: int, dPhiPerZ: float) -> None:
		self.m_nSamplesPerRound = int(nSamplesPerRound)
		self.m_dPhiPerZ = float(dPhiPerZ)

	def Apply(self, vecPt: Vector3Like) -> tuple[float, float, float]:
		pt = as_vec3(vecPt)
		fZ = float(pt[2])
		fRadius = VecOperations.fGetRadius(pt)
		fPhi = VecOperations.fGetPhi(pt) + self.m_dPhiPerZ * fZ
		return (fRadius, float(self.m_nSamplesPerRound * fPhi), fZ)


class CombinedTrafo(ICoordinateTrafo):
	def __init__(self, aTrafos: list[ICoordinateTrafo]) -> None:
		self.m_aTrafos = aTrafos

	def Apply(self, vecPt: Vector3Like) -> tuple[float, float, float]:
		fX, fY, fZ = as_vec3(vecPt)
		current: tuple[float, float, float] = (fX, fY, fZ)
		for xTrafo in self.m_aTrafos:
			current = xTrafo.Apply(current)
		return current


__all__ = [
	"ICoordinateTrafo",
	"ScaleTrafo",
	"FunctionalScaleTrafo",
	"RadialTrafo",
	"CombinedTrafo",
]



from __future__ import annotations


import ctypes

from _common.types import Vector3Like

from .._base import HandleOwner
from .._errors import PicoGKInvalidHandleError
from .._types import as_vec3

from .library import Library

class Lattice(HandleOwner):
    def __init__(self, handle: int | ctypes.c_void_p | None = None, *, owns_handle: bool = True) -> None:
        if handle is None:
            handle = Library._lib().Lattice_hCreate()
        if not handle:
            raise PicoGKInvalidHandleError("Lattice_hCreate returned null")
        super().__init__(handle, owns_handle=owns_handle)

    def _destroy_native(self) -> None:
        Library._lib().Lattice_Destroy(self._handle)

    def is_valid(self) -> bool:
        return bool(Library._lib().Lattice_bIsValid(self.handle))

    def add_sphere(self, center: Vector3Like, radius: float) -> "Lattice":
        self._ensure_open()
        c = as_vec3(center)
        Library._lib().Lattice_AddSphere(self.handle, ctypes.byref(c), ctypes.c_float(radius))
        return self

    AddSphere = add_sphere

    def add_beam(
        self,
        a: Vector3Like,
        b: Vector3Like,
        radius_a: float,
        radius_b: float,
        round_cap: bool = True,
    ) -> "Lattice":
        self._ensure_open()
        va = as_vec3(a)
        vb = as_vec3(b)
        Library._lib().Lattice_AddBeam(
            self.handle,
            ctypes.byref(va),
            ctypes.byref(vb),
            ctypes.c_float(radius_a),
            ctypes.c_float(radius_b),
            ctypes.c_bool(round_cap),
        )
        return self

    AddBeam = add_beam






from __future__ import annotations


class ListOperations:
	@staticmethod
	def aOverSampleList(aList: list[float], iSamplesPerStep: int) -> list[float]:
		out: list[float] = []
		for i in range(1, len(aList)):
			for j in range(int(iSamplesPerStep)):
				t = float(j) / float(iSamplesPerStep)
				out.append(float(aList[i - 1]) + t * (float(aList[i]) - float(aList[i - 1])))
		out.append(float(aList[-1]))
		return out

	@staticmethod
	def aSubsampleList(aList: list[float], iSampleSize: int) -> list[float]:
		out: list[float] = []
		step = max(1, int(iSampleSize))
		i = 0
		while i < len(aList):
			out.append(float(aList[i]))
			i += step
		if out[-1] != float(aList[-1]):
			out.append(float(aList[-1]))
		return out

	@staticmethod
	def iGetIndexOfMaxValue(aList: list[float]) -> int:
		max_val = float("-inf")
		idx = -1
		for i, v in enumerate(aList):
			if float(v) > max_val:
				max_val = float(v)
				idx = i
		return idx

	@staticmethod
	def iGetIndexOfMinValue(aList: list[float]) -> int:
		min_val = float("inf")
		idx = -1
		for i, v in enumerate(aList):
			if float(v) < min_val:
				min_val = float(v)
				idx = i
		return idx


__all__ = ["ListOperations"]



from .boundary_beam_thickness import BoundaryBeamThickness
from .cell_based_beam_thickness import CellBasedBeamThickness
from .constant_beam_thickness import ConstantBeamThickness
from .global_func_beam_thickness import GlobalFuncBeamThickness
from .i_beam_thickness import IBeamThickness

__all__ = [
	"IBeamThickness",
	"BoundaryBeamThickness",
	"CellBasedBeamThickness",
	"ConstantBeamThickness",
	"GlobalFuncBeamThickness",
]


from typing import Sequence

import numpy as np

from _common.types import Vector3Like


class Vector3Ext:
    @staticmethod
    def vecNormalized(vec: Vector3Like) -> tuple[float, float, float]:
        v = np.array(vec, dtype=np.float64)
        n = np.linalg.norm(v)
        if n <= 1e-12:
            return (0.0, 0.0, 0.0)
        v = v / n
        return (float(v[0]), float(v[1]), float(v[2]))

    @staticmethod
    def vecMirrored(vec: Vector3Like, vecPlanePoint: Vector3Like, vecPlaneNormalUnitVector: Vector3Like) -> tuple[float, float, float]:
        v = np.array(vec, dtype=np.float64)
        p = np.array(vecPlanePoint, dtype=np.float64)
        n = np.array(vecPlaneNormalUnitVector, dtype=np.float64)
        ln = np.linalg.norm(n)
        if ln <= 1e-12:
            raise ValueError("vecPlaneNormalUnitVector must be non-zero")
        n = n / ln
        out = v - (2.0 * np.dot(v - p, n) * n)
        return (float(out[0]), float(out[1]), float(out[2]))

    @staticmethod
    def vecTransformed(vec: Vector3Like, mat: Sequence[float]) -> tuple[float, float, float]:
        if len(mat) != 16:
            raise ValueError("mat must have 16 values (row-major 4x4)")
        m = np.array(mat, dtype=np.float64).reshape((4, 4))
        v = np.array((float(vec[0]), float(vec[1]), float(vec[2]), 1.0), dtype=np.float64)
        out = m @ v
        return (float(out[0]), float(out[1]), float(out[2]))





from __future__ import annotations

import math

from shape_kernel._types import Vector3Like, as_vec3

from picogk import IImplicit


class ImplicitSchwarzPrimitive(IImplicit):
	def __init__(self, fUnitSize: float, fWallThickness: float) -> None:
		self.m_fFrequencyScale = (2.0 * math.pi) / float(fUnitSize)
		self.m_fWallThickness = float(fWallThickness)

	def fSignedDistance(self, vecPt: Vector3Like) -> float:
		dX, dY, dZ = as_vec3(vecPt)
		dDist = (
			math.cos(self.m_fFrequencyScale * dX)
			+ math.cos(self.m_fFrequencyScale * dY)
			+ math.cos(self.m_fFrequencyScale * dZ)
		)
		return abs(dDist) - 0.5 * self.m_fWallThickness


__all__ = ["ImplicitSchwarzPrimitive"]


from __future__ import annotations


from typing import Callable, Sequence, TYPE_CHECKING

from picogk import Mesh, Voxels

from .._types import Vec3, Vector3Like

if TYPE_CHECKING:
    from ..frames.local_frames import LocalFrame
from .vec_operations import VecOperations


class MeshUtility:
    @staticmethod
    def mshFromGrid(aGrid: Sequence[Sequence[Vector3Like]]) -> Mesh:
        msh = Mesh()
        for i in range(1, len(aGrid)):
            for j in range(1, len(aGrid[i])):
                p1 = aGrid[i - 1][j - 1]
                p2 = aGrid[i - 1][j]
                p3 = aGrid[i][j]
                p4 = aGrid[i][j - 1]
                msh.nAddTriangle(p4, p1, p2)
                msh.nAddTriangle(p2, p3, p4)
        return msh

    @staticmethod
    def mshFromQuad(vecPt1: Vector3Like, vecPt2: Vector3Like, vecPt3: Vector3Like, vecPt4: Vector3Like) -> Mesh:
        msh = Mesh()
        msh.nAddTriangle(vecPt4, vecPt1, vecPt2)
        msh.nAddTriangle(vecPt2, vecPt3, vecPt4)
        return msh

    @staticmethod
    def AddQuad(msh: Mesh, vecPt1: Vector3Like, vecPt2: Vector3Like, vecPt3: Vector3Like, vecPt4: Vector3Like) -> None:
        msh.nAddTriangle(vecPt4, vecPt1, vecPt2)
        msh.nAddTriangle(vecPt2, vecPt3, vecPt4)

    @staticmethod
    def mshApplyTransformation(msh: Mesh, fnTrafo: Callable[[Vec3], Vec3]) -> Mesh:
        new_msh = Mesh()
        for i in range(msh.triangle_count()):
            a, b, c = msh.get_triangle_vertices(i)
            new_msh.nAddTriangle(fnTrafo(a), fnTrafo(b), fnTrafo(c))
        return new_msh

    @staticmethod
    def voxApplyTransformation(vox: Voxels, fnTrafo: Callable[[Vec3], Vec3]) -> Voxels:
        with Mesh.from_voxels(vox) as msh:
            new_msh = MeshUtility.mshApplyTransformation(msh, fnTrafo)
        return Voxels.from_mesh(new_msh)

    @staticmethod
    def mshTranslateMeshOntoFrame(msh: Mesh, oInputFrame: LocalFrame, oOutputFrame: LocalFrame) -> Mesh:
        def trafo(pt: Vec3) -> Vec3:
            rel = VecOperations.vecExpressPointInFrame(oInputFrame, pt)
            return VecOperations.vecTranslatePointOntoFrame(oOutputFrame, rel)
        return MeshUtility.mshApplyTransformation(msh, trafo)

    @staticmethod
    def Append(msh1: Mesh, msh2: Mesh) -> Mesh:
        return msh1.append(msh2)




from __future__ import annotations

import math

from shape_kernel import BaseBox, BaseLens, BasePipeSegment, LineModulation, LocalFrame, SurfaceModulation
from shape_kernel._types import Vec3

from ..unit_cells import CuboidCell, IUnitCell
from .i_cell_array import ICellArray


class ConformalCellArray(ICellArray):
	def __init__(self, oShape: BaseBox | BaseLens | BasePipeSegment, nNumberInX: int, nNumberInY: int, nNumberInZ: int) -> None:
		self.m_nNumberInX = int(nNumberInX)
		self.m_nNumberInY = int(nNumberInY)
		self.m_nNumberInZ = int(nNumberInZ)
		self.m_aUnitCells: list[IUnitCell] = []

		if isinstance(oShape, BaseBox):
			self._build_from_box(oShape)
		elif isinstance(oShape, BaseLens):
			self._build_from_lens(oShape)
		elif isinstance(oShape, BasePipeSegment):
			self._build_from_segment(oShape)
		else:
			raise TypeError("ConformalCellArray supports BaseBox, BaseLens, and BasePipeSegment only.")

	def _build_from_box(self, oBox: BaseBox) -> None:
		for nX in range(1, self.m_nNumberInX + 1):
			for nY in range(1, self.m_nNumberInY + 1):
				for nZ in range(1, self.m_nNumberInZ + 1):
					self.m_aUnitCells.append(
						CuboidCell(
							self.vecGetInternalBoxPt(oBox, nX - 1, nY - 1, nZ - 1),
							self.vecGetInternalBoxPt(oBox, nX, nY - 1, nZ - 1),
							self.vecGetInternalBoxPt(oBox, nX, nY, nZ - 1),
							self.vecGetInternalBoxPt(oBox, nX - 1, nY, nZ - 1),
							self.vecGetInternalBoxPt(oBox, nX - 1, nY - 1, nZ),
							self.vecGetInternalBoxPt(oBox, nX, nY - 1, nZ),
							self.vecGetInternalBoxPt(oBox, nX, nY, nZ),
							self.vecGetInternalBoxPt(oBox, nX - 1, nY, nZ),
						)
					)

	def _build_from_lens(self, oLens: BaseLens) -> None:
		for nX in range(1, self.m_nNumberInX + 1):
			for nY in range(1, self.m_nNumberInY + 1):
				for nZ in range(1, self.m_nNumberInZ + 1):
					self.m_aUnitCells.append(
						CuboidCell(
							self.vecGetInternalLensPt(oLens, nX - 1, nY - 1, nZ - 1),
							self.vecGetInternalLensPt(oLens, nX, nY - 1, nZ - 1),
							self.vecGetInternalLensPt(oLens, nX, nY, nZ - 1),
							self.vecGetInternalLensPt(oLens, nX - 1, nY, nZ - 1),
							self.vecGetInternalLensPt(oLens, nX - 1, nY - 1, nZ),
							self.vecGetInternalLensPt(oLens, nX, nY - 1, nZ),
							self.vecGetInternalLensPt(oLens, nX, nY, nZ),
							self.vecGetInternalLensPt(oLens, nX - 1, nY, nZ),
						)
					)

	def _build_from_segment(self, oSegment: BasePipeSegment) -> None:
		for nX in range(1, self.m_nNumberInX + 1):
			for nY in range(1, self.m_nNumberInY + 1):
				for nZ in range(1, self.m_nNumberInZ + 1):
					self.m_aUnitCells.append(
						CuboidCell(
							self.vecGetInternalSegmentPt(oSegment, nX - 1, nY - 1, nZ - 1),
							self.vecGetInternalSegmentPt(oSegment, nX, nY - 1, nZ - 1),
							self.vecGetInternalSegmentPt(oSegment, nX, nY, nZ - 1),
							self.vecGetInternalSegmentPt(oSegment, nX - 1, nY, nZ - 1),
							self.vecGetInternalSegmentPt(oSegment, nX - 1, nY - 1, nZ),
							self.vecGetInternalSegmentPt(oSegment, nX, nY - 1, nZ),
							self.vecGetInternalSegmentPt(oSegment, nX, nY, nZ),
							self.vecGetInternalSegmentPt(oSegment, nX - 1, nY, nZ),
						)
					)

	def vecGetInternalBoxPt(self, oBox: BaseBox, nX: int, nY: int, nZ: int) -> Vec3:
		fLengthRatio = float(nZ) / float(self.m_nNumberInZ)
		fWidthRatio = (2.0 * float(nX) / float(self.m_nNumberInX)) - 1.0
		fDepthRatio = (2.0 * float(nY) / float(self.m_nNumberInY)) - 1.0
		return oBox.vecGetSurfacePoint(fWidthRatio, fDepthRatio, fLengthRatio)

	def vecGetInternalLensPt(self, oLens: BaseLens, nX: int, nY: int, nZ: int) -> Vec3:
		fPhiRatio = float(nZ) / float(self.m_nNumberInZ)
		fHeightRatio = float(nX) / float(self.m_nNumberInX)
		fRadiusRatio = float(nY) / float(self.m_nNumberInY)
		return oLens.vecGetSurfacePoint(fHeightRatio, fPhiRatio, fRadiusRatio)

	def vecGetInternalSegmentPt(self, oSegment: BasePipeSegment, nX: int, nY: int, nZ: int) -> Vec3:
		fPhiRatio = float(nZ) / float(self.m_nNumberInZ)
		fLengthRatio = float(nX) / float(self.m_nNumberInX)
		fRadiusRatio = float(nY) / float(self.m_nNumberInY)
		return oSegment.vecGetSurfacePoint(fLengthRatio, fPhiRatio, fRadiusRatio)

	def aGetUnitCells(self) -> list[IUnitCell]:
		return self.m_aUnitCells


class ConformalShowcaseShapes:
	@staticmethod
	def oGetBox_01() -> BaseBox:
		oBox = BaseBox.from_frame(LocalFrame(), fLength=100.0)
		oBox.SetDepth(LineModulation(ConformalShowcaseShapes.fGetDepth_01))
		oBox.SetWidth(LineModulation(ConformalShowcaseShapes.fGetWidth_01))
		return oBox

	@staticmethod
	def fGetWidth_01(fLengthRatio: float) -> float:
		return 60.0 + 20.0 * math.cos(5.0 * float(fLengthRatio))

	@staticmethod
	def fGetDepth_01(fLengthRatio: float) -> float:
		return 80.0 - 40.0 * math.cos(3.0 * float(fLengthRatio))

	@staticmethod
	def oGetLens_01() -> BaseLens:
		oLens = BaseLens(LocalFrame(), 1.0, 20.0, 40.0)
		oLens.SetHeight(
			SurfaceModulation(ConformalShowcaseShapes.fGetLowerLens_01),
			SurfaceModulation(ConformalShowcaseShapes.fGetUpperLens_01),
		)
		return oLens

	@staticmethod
	def fGetLowerLens_01(fPhi: float, fRadiusRatio: float) -> float:
		return -20.0 + 20.0 * float(fRadiusRatio)

	@staticmethod
	def fGetUpperLens_01(fPhi: float, fRadiusRatio: float) -> float:
		return 20.0 + 5.0 * math.cos(2.0 * float(fRadiusRatio))

	@staticmethod
	def oGetSegment_01() -> BasePipeSegment:
		oSegment = BasePipeSegment.from_frame(
			LocalFrame(),
			100.0,
			20.0,
			40.0,
			LineModulation(0.0),
			LineModulation(ConformalShowcaseShapes.fGetPhiRange_01),
			BasePipeSegment.EMethod.MID_RANGE,
		)
		oSegment.SetRadius(
			SurfaceModulation(ConformalShowcaseShapes.fGetInnerRadius_01),
			SurfaceModulation(ConformalShowcaseShapes.fGetOuterRadius_01),
		)
		return oSegment

	@staticmethod
	def fGetPhiRange_01(fLengthRatio: float) -> float:
		return math.pi - 0.45 * math.pi * math.cos(3.0 * float(fLengthRatio))

	@staticmethod
	def fGetInnerRadius_01(fPhi: float, fLengthRatio: float) -> float:
		return 20.0 + 10.0 * float(fLengthRatio)

	@staticmethod
	def fGetOuterRadius_01(fPhi: float, fLengthRatio: float) -> float:
		return ConformalShowcaseShapes.fGetInnerRadius_01(fPhi, fLengthRatio) + 15.0 + 5.0 * math.cos(4.0 * float(fPhi))


__all__ = ["ConformalCellArray", "ConformalShowcaseShapes"]


from __future__ import annotations


import ctypes
import math
from typing import Sequence


from _common.types import Vector3Like

from .._base import HandleOwner
from .._errors import PicoGKInvalidHandleError
from .._types import ColorFloat, Vec3, as_vec3, vec3_tuple

from .library import Library

class PolyLine(HandleOwner):
    def __init__(self, color: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)) -> None:
        clr = ColorFloat(*[float(c) for c in color])
        h = Library._lib().PolyLine_hCreate(ctypes.byref(clr))
        if not h:
            raise PicoGKInvalidHandleError("PolyLine_hCreate returned null")
        super().__init__(h)
        self._bbox_min = [float("inf"), float("inf"), float("inf")]
        self._bbox_max = [float("-inf"), float("-inf"), float("-inf")]

    def _destroy_native(self) -> None:
        Library._lib().PolyLine_Destroy(self._handle)

    def add_vertex(self, xyz: Vector3Like) -> int:
        self._ensure_open()
        v = as_vec3(xyz)
        self._bbox_min[0] = min(self._bbox_min[0], float(v.x))
        self._bbox_min[1] = min(self._bbox_min[1], float(v.y))
        self._bbox_min[2] = min(self._bbox_min[2], float(v.z))
        self._bbox_max[0] = max(self._bbox_max[0], float(v.x))
        self._bbox_max[1] = max(self._bbox_max[1], float(v.y))
        self._bbox_max[2] = max(self._bbox_max[2], float(v.z))
        return int(Library._lib().PolyLine_nAddVertex(self.handle, ctypes.byref(v)))

    nAddVertex = add_vertex

    def Add(self, vertices: Sequence[Vector3Like]) -> "PolyLine":
        for vertex in vertices:
            self.add_vertex(vertex)
        return self

    def vertex_count(self) -> int:
        self._ensure_open()
        return int(Library._lib().PolyLine_nVertexCount(self.handle))

    nVertexCount = vertex_count

    def get_vertex(self, index: int) -> tuple[float, float, float]:
        self._ensure_open()
        v = Vec3()
        Library._lib().PolyLine_GetVertex(self.handle, int(index), ctypes.byref(v))
        return vec3_tuple(v)

    vecVertexAt = get_vertex

    def get_color(self) -> tuple[float, float, float, float]:
        clr = ColorFloat()
        Library._lib().PolyLine_GetColor(self.handle, ctypes.byref(clr))
        return (float(clr.r), float(clr.g), float(clr.b), float(clr.a))

    GetColor = get_color

    def bounding_box(self) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        if self.vertex_count() == 0:
            return ((float("inf"), float("inf"), float("inf")), (float("-inf"), float("-inf"), float("-inf")))
        return (
            (self._bbox_min[0], self._bbox_min[1], self._bbox_min[2]),
            (self._bbox_max[0], self._bbox_max[1], self._bbox_max[2]),
        )

    oBoundingBox = bounding_box

    def AddArrow(self, size_mm: float = 1.0, direction: Vector3Like | None = None) -> "PolyLine":
        if self.vertex_count() < 1:
            return self
        if direction is None:
            if self.vertex_count() < 2:
                return self
            start = self.get_vertex(self.vertex_count() - 2)
            end = self.get_vertex(self.vertex_count() - 1)
            dx = end[0] - start[0]
            dy = end[1] - start[1]
            dz = end[2] - start[2]
        else:
            dx, dy, dz = direction
        length = math.sqrt(dx * dx + dy * dy + dz * dz)
        if length <= 1e-12:
            return self
        dx /= length
        dy /= length
        dz /= length
        init = (1.0, 0.0, 0.0) if abs(dx) < 0.999 else (0.0, 1.0, 0.0)
        ux = dy * init[2] - dz * init[1]
        uy = dz * init[0] - dx * init[2]
        uz = dx * init[1] - dy * init[0]
        ul = math.sqrt(ux * ux + uy * uy + uz * uz)
        ux, uy, uz = ux / ul, uy / ul, uz / ul
        vx = dy * uz - dz * uy
        vy = dz * ux - dx * uz
        vz = dx * uy - dy * ux
        tip = self.get_vertex(self.vertex_count() - 1)
        base = (tip[0] - dx * size_mm, tip[1] - dy * size_mm, tip[2] - dz * size_mm)
        self.add_vertex((base[0] + ux * size_mm / 2, base[1] + uy * size_mm / 2, base[2] + uz * size_mm / 2))
        self.add_vertex((base[0] - ux * size_mm / 2, base[1] - uy * size_mm / 2, base[2] - uz * size_mm / 2))
        self.add_vertex(tip)
        self.add_vertex((base[0] + vx * size_mm / 2, base[1] + vy * size_mm / 2, base[2] + vz * size_mm / 2))
        self.add_vertex((base[0] - vx * size_mm / 2, base[1] - vy * size_mm / 2, base[2] - vz * size_mm / 2))
        self.add_vertex(tip)
        return self

    def AddCross(self, size_mm: float = 1.0) -> "PolyLine":
        if self.vertex_count() < 1:
            return self
        cx, cy, cz = self.get_vertex(self.vertex_count() - 1)
        self.add_vertex((cx + size_mm, cy, cz))
        self.add_vertex((cx - size_mm, cy, cz))
        self.add_vertex((cx, cy, cz))
        self.add_vertex((cx, cy + size_mm, cz))
        self.add_vertex((cx, cy - size_mm, cz))
        self.add_vertex((cx, cy, cz))
        self.add_vertex((cx, cy, cz + size_mm))
        self.add_vertex((cx, cy, cz - size_mm))
        self.add_vertex((cx, cy, cz))
        return self






from .animations import AnimGroupMatrixRotate, AnimViewRotate
from .keybindings import IKeyHandler, IViewerAction, KeyAction, KeyHandler, RotateToNextRoundAngleAction
from .runtime import go
from .vedo_viewer import VedoViewer

__all__ = [
    "IViewerAction",
    "IKeyHandler",
    "KeyAction",
    "KeyHandler",
    "RotateToNextRoundAngleAction",
    "AnimGroupMatrixRotate",
    "AnimViewRotate",
    "VedoViewer",
    "go",
]



from __future__ import annotations

from picogk import BBox3, Mesh, Voxels

from .._types import Vec3, Vector3Like, as_np3, as_vec3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from ..modulations import LineModulation
from .base_shape import BaseShape


class BaseBox(BaseShape):
	def __init__(
		self,
		aFrames: Frames,
		fWidth: float,
		fDepth: float,
		*,
		_length_steps: int,
	) -> None:
		super().__init__()
		self.m_aFrames = aFrames
		self.SetWidthSteps(5)
		self.SetDepthSteps(5)
		self.SetLengthSteps(_length_steps)
		self.m_oWidthModulation = LineModulation(fWidth)
		self.m_oDepthModulation = LineModulation(fDepth)

	@classmethod
	def from_frame(
		cls,
		oFrame: LocalFrame,
		fLength: float = 20.0,
		fWidth: float = 20.0,
		fDepth: float = 20.0,
	) -> "BaseBox":
		return cls(Frames.from_length(fLength, oFrame), fWidth, fDepth, _length_steps=5)

	@classmethod
	def from_frames(
		cls,
		aFrames: Frames,
		fWidth: float = 20.0,
		fDepth: float = 20.0,
	) -> "BaseBox":
		return cls(aFrames, fWidth, fDepth, _length_steps=500)

	@classmethod
	def from_bbox(
		cls,
		oBBox: BBox3 | tuple[Vector3Like, Vector3Like],
	) -> "BaseBox":
		if isinstance(oBBox, BBox3):
			box_min = (float(oBBox.min.x), float(oBBox.min.y), float(oBBox.min.z))
			box_max = (float(oBBox.max.x), float(oBBox.max.y), float(oBBox.max.z))
		else:
			box_min, box_max = oBBox

		bmin = as_np3(box_min)
		bmax = as_np3(box_max)
		centre = 0.5 * (bmin + bmax)
		centre[2] = bmin[2]
		return cls.from_frame(
			LocalFrame(as_vec3(centre), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0)),
			fLength=float(bmax[2] - bmin[2]),
			fWidth=float(bmax[0] - bmin[0]),
			fDepth=float(bmax[1] - bmin[1]),
		)

	def SetWidth(self, oModulation: LineModulation) -> None:
		self.m_oWidthModulation = oModulation
		self.SetWidthSteps(500)
		self.SetLengthSteps(500)

	def SetDepth(self, oModulation: LineModulation) -> None:
		self.m_oDepthModulation = oModulation
		self.SetDepthSteps(500)
		self.SetLengthSteps(500)

	def SetWidthSteps(self, nWidthSteps: int) -> None:
		self.m_nWidthSteps = max(5, int(nWidthSteps))

	def SetDepthSteps(self, nDepthSteps: int) -> None:
		self.m_nDepthSteps = max(5, int(nDepthSteps))

	def SetLengthSteps(self, nLengthSteps: int) -> None:
		self.m_nLengthSteps = max(5, int(nLengthSteps))

	def voxConstruct(self) -> Voxels:
		return Voxels.from_mesh(self.mshConstruct())

	def mshConstruct(self) -> Mesh:
		mesh = Mesh()
		self._add_top_surface(mesh, True)
		self._add_bottom_surface(mesh)
		self._add_front_surface(mesh, True)
		self._add_back_surface(mesh)
		self._add_right_surface(mesh, True)
		self._add_left_surface(mesh)
		return mesh

	def _add_rect(self, mesh: Mesh, pts: tuple[Vec3, Vec3, Vec3, Vec3], flip: bool) -> None:
		p0, p1, p2, p3 = pts
		if not flip:
			mesh.nAddTriangle(p0, p1, p2)
			mesh.nAddTriangle(p0, p2, p3)
		else:
			mesh.nAddTriangle(p0, p2, p1)
			mesh.nAddTriangle(p0, p3, p2)

	def _add_top_surface(self, mesh: Mesh, bFlip: bool = False) -> None:
		lr = self._length_ratio(self.m_nLengthSteps - 1)
		for iw in range(1, self.m_nWidthSteps):
			w1 = self._width_ratio(iw - 1)
			w2 = self._width_ratio(iw)
			for idepth in range(1, self.m_nDepthSteps):
				d1 = self._depth_ratio(idepth - 1)
				d2 = self._depth_ratio(idepth)
				self._add_rect(
					mesh,
					(
						self.vecGetSurfacePoint(w1, d1, lr),
						self.vecGetSurfacePoint(w1, d2, lr),
						self.vecGetSurfacePoint(w2, d2, lr),
						self.vecGetSurfacePoint(w2, d1, lr),
					),
					bFlip,
				)

	def _add_bottom_surface(self, mesh: Mesh, bFlip: bool = False) -> None:
		lr = self._length_ratio(0)
		for iw in range(1, self.m_nWidthSteps):
			w1 = self._width_ratio(iw - 1)
			w2 = self._width_ratio(iw)
			for idepth in range(1, self.m_nDepthSteps):
				d1 = self._depth_ratio(idepth - 1)
				d2 = self._depth_ratio(idepth)
				self._add_rect(
					mesh,
					(
						self.vecGetSurfacePoint(w1, d1, lr),
						self.vecGetSurfacePoint(w1, d2, lr),
						self.vecGetSurfacePoint(w2, d2, lr),
						self.vecGetSurfacePoint(w2, d1, lr),
					),
					bFlip,
				)

	def _add_front_surface(self, mesh: Mesh, bFlip: bool = False) -> None:
		wr = self._width_ratio(0)
		for il in range(1, self.m_nLengthSteps):
			lr1 = self._length_ratio(il - 1)
			lr2 = self._length_ratio(il)
			for idepth in range(1, self.m_nDepthSteps):
				d1 = self._depth_ratio(idepth - 1)
				d2 = self._depth_ratio(idepth)
				self._add_rect(mesh, (self.vecGetSurfacePoint(wr, d1, lr1), self.vecGetSurfacePoint(wr, d2, lr1), self.vecGetSurfacePoint(wr, d2, lr2), self.vecGetSurfacePoint(wr, d1, lr2)), bFlip)

	def _add_back_surface(self, mesh: Mesh, bFlip: bool = False) -> None:
		wr = self._width_ratio(self.m_nWidthSteps - 1)
		for il in range(1, self.m_nLengthSteps):
			lr1 = self._length_ratio(il - 1)
			lr2 = self._length_ratio(il)
			for idepth in range(1, self.m_nDepthSteps):
				d1 = self._depth_ratio(idepth - 1)
				d2 = self._depth_ratio(idepth)
				self._add_rect(mesh, (self.vecGetSurfacePoint(wr, d1, lr1), self.vecGetSurfacePoint(wr, d2, lr1), self.vecGetSurfacePoint(wr, d2, lr2), self.vecGetSurfacePoint(wr, d1, lr2)), bFlip)

	def _add_right_surface(self, mesh: Mesh, bFlip: bool = False) -> None:
		dr = self._depth_ratio(self.m_nDepthSteps - 1)
		for il in range(1, self.m_nLengthSteps):
			lr1 = self._length_ratio(il - 1)
			lr2 = self._length_ratio(il)
			for iw in range(1, self.m_nWidthSteps):
				w1 = self._width_ratio(iw - 1)
				w2 = self._width_ratio(iw)
				self._add_rect(mesh, (self.vecGetSurfacePoint(w1, dr, lr1), self.vecGetSurfacePoint(w2, dr, lr1), self.vecGetSurfacePoint(w2, dr, lr2), self.vecGetSurfacePoint(w1, dr, lr2)), bFlip)

	def _add_left_surface(self, mesh: Mesh, bFlip: bool = False) -> None:
		dr = self._depth_ratio(0)
		for il in range(1, self.m_nLengthSteps):
			lr1 = self._length_ratio(il - 1)
			lr2 = self._length_ratio(il)
			for iw in range(1, self.m_nWidthSteps):
				w1 = self._width_ratio(iw - 1)
				w2 = self._width_ratio(iw)
				self._add_rect(mesh, (self.vecGetSurfacePoint(w1, dr, lr1), self.vecGetSurfacePoint(w2, dr, lr1), self.vecGetSurfacePoint(w2, dr, lr2), self.vecGetSurfacePoint(w1, dr, lr2)), bFlip)

	def _width_ratio(self, step: int) -> float:
		return -1.0 + (2.0 / float(self.m_nWidthSteps - 1)) * float(step)

	def _depth_ratio(self, step: int) -> float:
		return -1.0 + (2.0 / float(self.m_nDepthSteps - 1)) * float(step)

	def _length_ratio(self, step: int) -> float:
		return (1.0 / float(self.m_nLengthSteps - 1)) * float(step)

	def fGetWidth(self, fLengthRatio: float) -> float:
		return self.m_oWidthModulation.fGetModulation(fLengthRatio)

	def fGetDepth(self, fLengthRatio: float) -> float:
		return self.m_oDepthModulation.fGetModulation(fLengthRatio)

	def vecGetSurfacePoint(self, fWidthRatio: float, fDepthRatio: float, fLengthRatio: float) -> Vec3:
		spine = as_np3(self.m_aFrames.vecGetSpineAlongLength(fLengthRatio))
		lx = as_np3(self.m_aFrames.vecGetLocalXAlongLength(fLengthRatio))
		ly = as_np3(self.m_aFrames.vecGetLocalYAlongLength(fLengthRatio))
		x = 0.5 * fWidthRatio * self.fGetWidth(fLengthRatio)
		y = 0.5 * fDepthRatio * self.fGetDepth(fLengthRatio)
		return self.m_fnTrafo(as_vec3(spine + x * lx + y * ly))


__all__ = ["BaseBox"]



from __future__ import annotations

import math

import numpy as np
from picogk import Lattice

from .._types import Vector3Like, as_np3, as_vec3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from .lattice_pipe import LatticePipe


class LatticeManifold(LatticePipe):
	def __init__(
		self,
		aFrames: Frames,
		fRadius: float,
		fMaxOverhangAngle: float,
		bExtendBothSides: bool,
		fMinPrintableRadius: float,
		*,
		_length_steps: int,
	) -> None:
		super().__init__(aFrames, fRadius, _length_steps=_length_steps)
		self.m_fMaxPrintableRadius = fMinPrintableRadius
		self.m_fLimitAngle = fMaxOverhangAngle
		self.m_bExtendBothSides = bExtendBothSides

	@classmethod
	def from_frame(
		cls,
		oFrame: LocalFrame,
		fLength: float = 20.0,
		fRadius: float = 10.0,
		fMaxOverhangAngle: float = 45.0,
		bExtendBothSides: bool = False,
		fMinPrintableRadius: float = 0.1,
	) -> "LatticeManifold":
		return cls(
			Frames.from_length(fLength, oFrame),
			fRadius,
			fMaxOverhangAngle,
			bExtendBothSides,
			fMinPrintableRadius,
			_length_steps=100,
		)

	@classmethod
	def from_frames(
		cls,
		aFrames: Frames,
		fRadius: float = 10.0,
		fMaxOverhangAngle: float = 45.0,
		bExtendBothSides: bool = False,
		fMinPrintableRadius: float = 0.1,
	) -> "LatticeManifold":
		return cls(
			aFrames,
			fRadius,
			fMaxOverhangAngle,
			bExtendBothSides,
			fMinPrintableRadius,
			_length_steps=500,
		)

	def latConstruct(self) -> Lattice:
		lattice = Lattice()
		for iz in range(self.m_nLengthSteps):
			lr = (1.0 / self.m_nLengthSteps) * iz
			p = self.vecGetSpinePoint(lr)
			beam = self.fGetRadius(lr)
			lattice.add_beam(p, p, beam, beam)
			self._add_tip(lattice, p, beam, True)
			if self.m_bExtendBothSides:
				self._add_tip(lattice, p, beam, False)
		return lattice

	def _add_tip(self, lattice: Lattice, vecPt: Vector3Like, fBeam: float, bZPositive: bool = True) -> None:
		half_alpha = 90.0 - self.m_fLimitAngle
		r = fBeam
		h = r * (1.0 - math.cos(math.radians(half_alpha)))
		s = 2.0 * r * math.sin(math.radians(half_alpha))
		tip_length = math.tan(math.radians(half_alpha)) * (0.5 * s - self.m_fMaxPrintableRadius)
		p = as_np3(vecPt)

		if bZPositive:
			mid = p + (r - h) * np.array([0.0, 0.0, 1.0], dtype=np.float64)
			tip = mid + tip_length * np.array([0.0, 0.0, 1.0], dtype=np.float64)
		else:
			mid = p - (r - h) * np.array([0.0, 0.0, 1.0], dtype=np.float64)
			tip = mid - tip_length * np.array([0.0, 0.0, 1.0], dtype=np.float64)

		lattice.add_beam(as_vec3(mid), as_vec3(tip), 0.5 * s, self.m_fMaxPrintableRadius, False)


__all__ = ["LatticeManifold"]




from enum import IntEnum
import math
from typing import Sequence

import numpy as np

from _common.types import ColorLike, clamp01
from .._common import to_rgba


class Image:
    class EType(IntEnum):
        BW = 0
        GRAY = 1
        COLOR = 2

    def __init__(self, width: int, height: int, eType: "Image.EType") -> None:
        if width < 1 or height < 1:
            raise ValueError("Image width and height must be larger than 0")
        self.nWidth = int(width)
        self.nHeight = int(height)
        self.eType = eType

    def clrValue(self, x: int, y: int) -> tuple[float, float, float, float]:
        if not (0 <= x < self.nWidth and 0 <= y < self.nHeight):
            return (0.0, 0.0, 0.0, 1.0)
        data = getattr(self, "_data", None)
        if data is None:
            f = self.fValue(x, y)
            return (f, f, f, 1.0)
        if self.eType == Image.EType.COLOR and len(data.shape) == 3 and data.shape[2] >= 4:
            px = data[y, x, :]
            return (float(px[0]), float(px[1]), float(px[2]), float(px[3]))
        f = float(data[y, x])
        return (f, f, f, 1.0)

    def fValue(self, x: int, y: int) -> float:
        if not (0 <= x < self.nWidth and 0 <= y < self.nHeight):
            return 0.0
        data = getattr(self, "_data", None)
        if data is None:
            clr = self.clrValue(x, y)
            return float((clr[0] + clr[1] + clr[2]) / 3.0)
        if self.eType == Image.EType.COLOR and len(data.shape) == 3 and data.shape[2] >= 3:
            px = data[y, x, :]
            return float((px[0] + px[1] + px[2]) / 3.0)
        if self.eType == Image.EType.BW:
            return 1.0 if bool(data[y, x]) else 0.0
        return float(data[y, x])

    def bValue(self, x: int, y: int) -> bool:
        return self.fValue(x, y) > 0.5

    def SetValue(self, x: int, y: int, value: bool | float | ColorLike) -> None:
        if not (0 <= x < self.nWidth and 0 <= y < self.nHeight):
            return
        data = getattr(self, "_data", None)
        if data is None:
            raise RuntimeError("Image subclass must define _data or override SetValue")
        if self.eType == Image.EType.BW:
            if isinstance(value, bool):
                data[y, x] = value
            elif isinstance(value, (float, int)):
                data[y, x] = float(value) > 0.5
            else:
                rgba = to_rgba(value)
                data[y, x] = ((rgba[0] + rgba[1] + rgba[2]) / 3.0) > 0.5
            return
        if self.eType == Image.EType.GRAY:
            if isinstance(value, bool):
                data[y, x] = 1.0 if value else 0.0
            elif isinstance(value, (float, int)):
                data[y, x] = float(value)
            else:
                rgba = to_rgba(value)
                data[y, x] = (rgba[0] + rgba[1] + rgba[2]) / 3.0
            return
        if isinstance(value, bool):
            c = 1.0 if value else 0.0
            data[y, x, :] = np.array([c, c, c, 1.0], dtype=np.float32)
        elif isinstance(value, (float, int)):
            c = float(value)
            data[y, x, :] = np.array([c, c, c, 1.0], dtype=np.float32)
        else:
            data[y, x, :] = np.array(to_rgba(value), dtype=np.float32)

    def byGetValue(self, x: int, y: int) -> int:
        return int(round(clamp01(self.fValue(x, y)) * 255.0))

    def SetValueByte(self, x: int, y: int, byValue: int) -> None:
        self.SetValue(x, y, float(int(byValue) & 0xFF) / 255.0)

    def sGetBgr24(self, x: int, y: int) -> tuple[int, int, int]:
        r, g, b, _ = self.clrValue(x, y)
        return (
            int(round(clamp01(b) * 255.0)),
            int(round(clamp01(g) * 255.0)),
            int(round(clamp01(r) * 255.0)),
        )

    def SetBgr24(self, x: int, y: int, clr: Sequence[int] | ColorLike) -> None:
        if len(clr) < 3:
            raise ValueError("BGR color must contain at least 3 channels")
        b = float(clr[0])
        g = float(clr[1])
        r = float(clr[2])
        if max(abs(b), abs(g), abs(r)) > 1.0:
            self.SetValue(x, y, (r / 255.0, g / 255.0, b / 255.0, 1.0))
        else:
            self.SetValue(x, y, (r, g, b, 1.0))

    def sGetBgra32(self, x: int, y: int) -> tuple[int, int, int, int]:
        r, g, b, a = self.clrValue(x, y)
        return (
            int(round(clamp01(b) * 255.0)),
            int(round(clamp01(g) * 255.0)),
            int(round(clamp01(r) * 255.0)),
            int(round(clamp01(a) * 255.0)),
        )

    def SetBgra32(self, x: int, y: int, clr: Sequence[int] | ColorLike) -> None:
        if len(clr) < 4:
            raise ValueError("BGRA color must contain 4 channels")
        b = float(clr[0])
        g = float(clr[1])
        r = float(clr[2])
        a = float(clr[3])
        if max(abs(b), abs(g), abs(r), abs(a)) > 1.0:
            self.SetValue(x, y, (r / 255.0, g / 255.0, b / 255.0, a / 255.0))
        else:
            self.SetValue(x, y, (r, g, b, a))

    def sGetRgb24(self, x: int, y: int) -> tuple[int, int, int]:
        r, g, b, _ = self.clrValue(x, y)
        return (
            int(round(clamp01(r) * 255.0)),
            int(round(clamp01(g) * 255.0)),
            int(round(clamp01(b) * 255.0)),
        )

    def sGetRgba32(self, x: int, y: int) -> tuple[int, int, int, int]:
        r, g, b, a = self.clrValue(x, y)
        return (
            int(round(clamp01(r) * 255.0)),
            int(round(clamp01(g) * 255.0)),
            int(round(clamp01(b) * 255.0)),
            int(round(clamp01(a) * 255.0)),
        )

    def SetRgb24(self, x: int, y: int, clr: Sequence[int] | ColorLike) -> None:
        if len(clr) < 3:
            raise ValueError("RGB color must contain at least 3 channels")
        r = float(clr[0])
        g = float(clr[1])
        b = float(clr[2])
        if max(abs(r), abs(g), abs(b)) > 1.0:
            self.SetValue(x, y, (r / 255.0, g / 255.0, b / 255.0, 1.0))
        else:
            self.SetValue(x, y, (r, g, b, 1.0))

    def SetRgba32(self, x: int, y: int, clr: Sequence[int] | ColorLike) -> None:
        if len(clr) < 4:
            raise ValueError("RGBA color must contain 4 channels")
        r = float(clr[0])
        g = float(clr[1])
        b = float(clr[2])
        a = float(clr[3])
        if max(abs(r), abs(g), abs(b), abs(a)) > 1.0:
            self.SetValue(x, y, (r / 255.0, g / 255.0, b / 255.0, a / 255.0))
        else:
            self.SetValue(x, y, (r, g, b, a))

    def clrGetAtNormalized(self, tx: float, ty: float) -> tuple[float, float, float, float]:
        fx = (tx * self.nWidth) - 1.0
        fy = (ty * self.nHeight) - 1.0
        x0 = max(0, min(self.nWidth - 1, int(math.floor(fx))))
        y0 = max(0, min(self.nHeight - 1, int(math.floor(fy))))
        x1 = max(0, min(self.nWidth - 1, x0 + 1))
        y1 = max(0, min(self.nHeight - 1, y0 + 1))
        dx = fx - x0
        dy = fy - y0
        c00 = self.clrValue(x0, y0)
        c10 = self.clrValue(x1, y0)
        c01 = self.clrValue(x0, y1)
        c11 = self.clrValue(x1, y1)

        def lerp(ca: tuple[float, float, float, float], cb: tuple[float, float, float, float], t: float) -> tuple[float, float, float, float]:
            return tuple(float(ca[i] + (cb[i] - ca[i]) * t) for i in range(4))  # type: ignore[return-value]

        c0 = lerp(c00, c10, dx)
        c1 = lerp(c01, c11, dx)
        return lerp(c0, c1, dy)

    def DrawLine(self, x0: int, y0: int, x1: int, y1: int, value: bool | float | ColorLike) -> None:
        dx = abs(x1 - x0)
        dy = abs(y1 - y0)
        sx = 1 if x0 < x1 else -1
        sy = 1 if y0 < y1 else -1
        err = dx - dy
        while True:
            self.SetValue(x0, y0, value)
            if x0 == x1 and y0 == y1:
                break
            e2 = 2 * err
            if e2 > -dy:
                err -= dy
                x0 += sx
            if e2 < dx:
                err += dx
                y0 += sy


class ImageBW(Image):
    def __init__(self, width: int, height: int) -> None:
        super().__init__(width, height, Image.EType.BW)
        self._data = np.zeros((height, width), dtype=np.bool_)

    def _in_bounds(self, x: int, y: int) -> bool:
        return 0 <= x < self.nWidth and 0 <= y < self.nHeight

    def bValue(self, x: int, y: int) -> bool:
        if not self._in_bounds(x, y):
            return False
        return bool(self._data[y, x])

    def fValue(self, x: int, y: int) -> float:
        return 1.0 if self.bValue(x, y) else 0.0

    def clrValue(self, x: int, y: int) -> tuple[float, float, float, float]:
        f = self.fValue(x, y)
        return (f, f, f, 1.0)

    def SetValue(self, x: int, y: int, value: bool | float | ColorLike) -> None:
        if not self._in_bounds(x, y):
            return
        if isinstance(value, bool):
            self._data[y, x] = value
            return
        if isinstance(value, (float, int)):
            self._data[y, x] = float(value) > 0.5
            return
        rgba = to_rgba(value)
        self._data[y, x] = ((rgba[0] + rgba[1] + rgba[2]) / 3.0) > 0.5


class ImageGrayScale(Image):
    def __init__(self, width: int, height: int) -> None:
        super().__init__(width, height, Image.EType.GRAY)
        self._data = np.zeros((height, width), dtype=np.float32)

    def _in_bounds(self, x: int, y: int) -> bool:
        return 0 <= x < self.nWidth and 0 <= y < self.nHeight

    def fValue(self, x: int, y: int) -> float:
        if not self._in_bounds(x, y):
            return 0.0
        return float(self._data[y, x])

    def bValue(self, x: int, y: int) -> bool:
        return self.fValue(x, y) > 0.5

    def clrValue(self, x: int, y: int) -> tuple[float, float, float, float]:
        f = self.fValue(x, y)
        return (f, f, f, 1.0)

    def SetValue(self, x: int, y: int, value: bool | float | ColorLike) -> None:
        if not self._in_bounds(x, y):
            return
        if isinstance(value, bool):
            self._data[y, x] = 1.0 if value else 0.0
            return
        if isinstance(value, (float, int)):
            self._data[y, x] = float(value)
            return
        rgba = to_rgba(value)
        self._data[y, x] = (rgba[0] + rgba[1] + rgba[2]) / 3.0

    def bContainsActivePixels(self, fThreshold: float = 0.0) -> bool:
        return bool(np.any(self._data <= float(fThreshold)))


class ImageColor(Image):
    def __init__(self, width: int, height: int) -> None:
        super().__init__(width, height, Image.EType.COLOR)
        self._data = np.zeros((height, width, 4), dtype=np.float32)
        self._data[:, :, 3] = 1.0

    def _in_bounds(self, x: int, y: int) -> bool:
        return 0 <= x < self.nWidth and 0 <= y < self.nHeight

    def clrValue(self, x: int, y: int) -> tuple[float, float, float, float]:
        if not self._in_bounds(x, y):
            return (0.0, 0.0, 0.0, 1.0)
        px = self._data[y, x, :]
        return (float(px[0]), float(px[1]), float(px[2]), float(px[3]))

    def fValue(self, x: int, y: int) -> float:
        r, g, b, _ = self.clrValue(x, y)
        return (r + g + b) / 3.0

    def bValue(self, x: int, y: int) -> bool:
        return self.fValue(x, y) > 0.5

    def SetValue(self, x: int, y: int, value: bool | float | ColorLike) -> None:
        if not self._in_bounds(x, y):
            return
        if isinstance(value, bool):
            c = 1.0 if value else 0.0
            self._data[y, x, :] = np.array([c, c, c, 1.0], dtype=np.float32)
            return
        if isinstance(value, (float, int)):
            c = float(value)
            self._data[y, x, :] = np.array([c, c, c, 1.0], dtype=np.float32)
            return
        self._data[y, x, :] = np.array(to_rgba(value), dtype=np.float32)


class ImageRgb24(ImageColor):
    def __init__(self, width_or_source: int | Image, height: int | None = None) -> None:
        if isinstance(width_or_source, Image):
            src = width_or_source
            super().__init__(src.nWidth, src.nHeight)
            for x in range(self.nWidth):
                for y in range(self.nHeight):
                    self.SetRgb24(x, y, src.sGetRgb24(x, y))
            return
        if height is None:
            raise ValueError("height is required when width is an integer")
        super().__init__(int(width_or_source), int(height))

    def clrValue(self, x: int, y: int) -> tuple[float, float, float, float]:
        r, g, b, _ = super().clrValue(x, y)
        return (r, g, b, 1.0)

    def SetValue(self, x: int, y: int, value: bool | float | ColorLike) -> None:
        if isinstance(value, Sequence) and len(value) >= 3:
            super().SetValue(x, y, (float(value[0]), float(value[1]), float(value[2]), 1.0))
            return
        super().SetValue(x, y, value)
        if 0 <= x < self.nWidth and 0 <= y < self.nHeight:
            self._data[y, x, 3] = 1.0


class ImageRgba32(ImageColor):
    def __init__(self, width_or_source: int | Image, height: int | None = None) -> None:
        if isinstance(width_or_source, Image):
            src = width_or_source
            super().__init__(src.nWidth, src.nHeight)
            for x in range(self.nWidth):
                for y in range(self.nHeight):
                    self.SetRgba32(x, y, src.sGetRgba32(x, y))
            return
        if height is None:
            raise ValueError("height is required when width is an integer")
        super().__init__(int(width_or_source), int(height))



from __future__ import annotations

import math
from collections.abc import Callable

import numpy as np

from picogk import Library, Mesh, VedoViewer

from .._types import Vec3, as_np3
from .color_scale_2d import IColorScale


class MeshPainter:
	ColorScaleFunc = Callable[[Vec3, Vec3, Vec3], float]

	@staticmethod
	def _build_submesh_classes(oMesh: Mesh, nClasses: int) -> list[Mesh]:
		return [Mesh() for _ in range(int(nClasses))]

	@staticmethod
	def _emit_preview_groups(
		segments: list[tuple[Mesh, tuple[float, float, float, float]]],
		oViewer: VedoViewer | None = None,
	) -> None:
		viewer = oViewer if oViewer is not None else Library.oViewer()
		if viewer is None:
			return
		from ..functions import Sh

		for msh, clr in segments:
			if msh.nTriangleCount() <= 0:
				continue
			Sh.PreviewMesh(msh, clr, oViewer=viewer)

	@staticmethod
	def PreviewOverhangAngle(
		oMesh: Mesh,
		xScale: IColorScale,
		bShowOnlyDownFacing: bool,
		nClasses: int = 30,
		oViewer: VedoViewer | None = None,
	) -> list[tuple[Mesh, tuple[float, float, float, float]]]:
		sub_meshes = MeshPainter._build_submesh_classes(oMesh, nClasses)
		fMinAngle = xScale.fGetMinValue()
		fMaxAngle = xScale.fGetMaxValue()
		dAngle = (fMaxAngle - fMinAngle) / (nClasses - 1.0)

		n_triangles = oMesh.nTriangleCount()
		for i in range(n_triangles):
			vecA, vecB, vecC = oMesh.GetTriangle(i)
			a = as_np3(vecA)
			b = as_np3(vecB)
			c = as_np3(vecC)
			normal = np.cross(a - b, c - b)
			n_norm = float(np.linalg.norm(normal))
			if n_norm <= 1e-12:
				continue
			normal = normal / n_norm

			dR = math.sqrt(float(normal[0] * normal[0] + normal[1] * normal[1]))
			dZ = abs(float(normal[2]))
			overhang = math.atan2(dZ, dR) / math.pi * 180.0
			overhang = min(max(overhang, 0.0), 90.0)

			if bShowOnlyDownFacing and normal[2] < 0.0:
				overhang = 0.0

			ratio = (overhang - fMinAngle) / (fMaxAngle - fMinAngle)
			index = int(min(max(ratio, 0.0), 1.0) * (nClasses - 1))
			sub_meshes[index].nAddTriangle(vecA, vecB, vecC)

		segments = [(sub_meshes[i], xScale.clrGetColor(fMinAngle + i * dAngle)) for i in range(nClasses)]
		MeshPainter._emit_preview_groups(segments, oViewer)
		return segments

	@staticmethod
	def PreviewCustomProperty(
		oMesh: Mesh,
		xScale: IColorScale,
		oColorFunc: ColorScaleFunc,
		nClasses: int = 30,
		oViewer: VedoViewer | None = None,
	) -> list[tuple[Mesh, tuple[float, float, float, float]]]:
		sub_meshes = MeshPainter._build_submesh_classes(oMesh, nClasses)
		fMinValue = xScale.fGetMinValue()
		fMaxValue = xScale.fGetMaxValue()
		dValue = (fMaxValue - fMinValue) / (nClasses - 1.0)

		n_triangles = oMesh.nTriangleCount()
		for i in range(n_triangles):
			vecA, vecB, vecC = oMesh.GetTriangle(i)
			value = float(oColorFunc(vecA, vecB, vecC))
			ratio = min(max((value - fMinValue) / (fMaxValue - fMinValue), 0.0), 1.0)
			index = int(ratio * (nClasses - 1))
			sub_meshes[index].nAddTriangle(vecA, vecB, vecC)

		segments = [(sub_meshes[i], xScale.clrGetColor(fMinValue + i * dValue)) for i in range(nClasses)]
		MeshPainter._emit_preview_groups(segments, oViewer)
		return segments

	@staticmethod
	def PreviewCustomDeformation(
		oMesh: Mesh,
		xScale: IColorScale,
		oColorFunc: ColorScaleFunc,
		fnTrafo: Callable[[Vec3], Vec3],
		nClasses: int = 30,
		oViewer: VedoViewer | None = None,
	) -> list[tuple[Mesh, tuple[float, float, float, float]]]:
		sub_meshes = MeshPainter._build_submesh_classes(oMesh, nClasses)
		fMinValue = xScale.fGetMinValue()
		fMaxValue = xScale.fGetMaxValue()
		dValue = (fMaxValue - fMinValue) / (nClasses - 1.0)

		n_triangles = oMesh.nTriangleCount()
		for i in range(n_triangles):
			vecA, vecB, vecC = oMesh.GetTriangle(i)
			value = float(oColorFunc(vecA, vecB, vecC))
			ratio = min(max((value - fMinValue) / (fMaxValue - fMinValue), 0.0), 1.0)
			index = int(ratio * (nClasses - 1))
			sub_meshes[index].nAddTriangle(fnTrafo(vecA), fnTrafo(vecB), fnTrafo(vecC))

		segments = [(sub_meshes[i], xScale.clrGetColor(fMinValue + i * dValue)) for i in range(nClasses)]
		MeshPainter._emit_preview_groups(segments, oViewer)
		return segments


__all__ = ["MeshPainter"]
